import argparse
import pathlib
from typing import Any

import pandas as pd

from grading_tools.common.commands import CommandModule
from grading_tools.common.defaults import NamingDictionary
from grading_tools.common.excel_utils import write_to_excel, write_table_helper
from grading_tools.common.utils import read_moodle_students_csv


def find_group_index(full_groups_lists_df: pd.DataFrame, group_name: str) -> int:
    matching_cols = full_groups_lists_df[full_groups_lists_df == group_name]
    return matching_cols.index[0] if len(matching_cols) > 0 else pd.NA


def gen_group_info_table(output_file: str, moodle_students_list: str, group_formats: list[str],
                         naming_dictionary: NamingDictionary, grouping_names: list[str] | None = None, **config):
    moodle_df = read_moodle_students_csv(moodle_students_list)

    moodle_df['GroupsList'] = moodle_df[naming_dictionary.MOODLE_GROUPS_GROUPS_COL].map(
        lambda s: [str.strip(g) for g in s.split(',')],
        na_action='ignore')
    exploded = moodle_df.explode('GroupsList', ignore_index=True)

    groupings = list(
        zip(grouping_names, group_formats) if grouping_names is not None else ((f'Grouping {i}', f) for i, f in
                                                                               enumerate(group_formats,
                                                                                         start=1)))
    full_groups_lists_df = pd.DataFrame(
        {g: [str.format(f, j) for j in range(0, 1000)] for g, f in groupings},
        index=range(0, 1000))

    grouping_info_cols = []
    for grouping in full_groups_lists_df.columns:
        all_groups_in_grouping = full_groups_lists_df[grouping]
        exploded[grouping] = exploded['GroupsList'].where(exploded['GroupsList'].isin(all_groups_in_grouping),
                                                          inplace=False)
        exploded[grouping + ' ID'] = exploded[grouping].map(lambda v: find_group_index(all_groups_in_grouping, v))
        grouping_info_cols.extend([grouping, grouping + ' ID'])

    # joining back together

    tables = [
        moodle_df.set_index(naming_dictionary.MOODLE_GROUPS_MATR_COL)[
            list(naming_dictionary.MOODLE_GROUPS_IDENTITY_COLS)]]
    for grouping in full_groups_lists_df.columns:
        tables.append(
            exploded.loc[exploded[grouping].notna(), [naming_dictionary.MOODLE_GROUPS_MATR_COL, grouping,
                                                      grouping + ' ID']].set_index(
                naming_dictionary.MOODLE_GROUPS_MATR_COL))
    group_info_df = pd.concat(tables, axis='columns', ignore_index=False).reset_index(drop=False)
    group_info_df.index.name = 'No'
    group_info_df.columns = [naming_dictionary.MATR_COL, naming_dictionary.FIRST_NAME_COL,
                             naming_dictionary.LAST_NAME_COL] + grouping_info_cols
    group_info_df = pd.DataFrame(group_info_df, columns=list(naming_dictionary.IDENTITY_COLS) + [
        naming_dictionary.MATR_COL] + grouping_info_cols)

    def my_write(exwr: pd.ExcelWriter):
        group_info_table_name = naming_dictionary.GROUPS_INFO_TABLE_NAME
        if config['verbose']:
            print(group_info_df)
        write_table_helper(exwr, group_info_df, group_info_table_name)

        for grouping, _ in groupings:
            grouping_table = group_info_df[group_info_df[grouping].notna()][
                list(naming_dictionary.IDENTITY_COLS) + [naming_dictionary.MATR_COL] + [grouping,
                                                                                        grouping + ' ID']].reset_index(
                drop=True)
            grouping_table.columns = list(naming_dictionary.IDENTITY_COLS) + [naming_dictionary.MATR_COL] + [
                naming_dictionary.GRADING_GROUP_NAME_COL, naming_dictionary.GRADING_GROUP_ID_COL]
            grouping_table.index.name = 'No'
            grouping_table_name = grouping.replace(' ', '_')
            if config['verbose']:
                print(grouping_table)
            write_table_helper(exwr, grouping_table, grouping_table_name)

            groups_df = grouping_table[
                [naming_dictionary.GRADING_GROUP_NAME_COL, naming_dictionary.GRADING_GROUP_ID_COL]]
            groups_df = groups_df.groupby(
                [naming_dictionary.GRADING_GROUP_NAME_COL, naming_dictionary.GRADING_GROUP_ID_COL],
                group_keys=True).agg(**{'Member Count': pd.NamedAgg(naming_dictionary.GRADING_GROUP_NAME_COL, 'count')})
            groups_df = groups_df.reset_index(drop=False).sort_values(naming_dictionary.GRADING_GROUP_ID_COL,
                                                                      ascending=True)
            groups_df.index.name = 'No'
            grouping_group_list_table_name = 'Groups_' + grouping.replace(' ', '_')
            if config['verbose']:
                print(groups_df)
            write_table_helper(exwr, groups_df, grouping_group_list_table_name)

    path = pathlib.Path(output_file)
    if path.is_dir():
        print(path)
        path = path.joinpath(config['DEFAULT_GROUPS_FILE_NAME'])
    write_to_excel(path, my_write, err_on_file_existence=False)


def register_gen_group_info(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults):
    parser.add_argument('-o', '--output-file', required=True, default=defaults['DEFAULT_GROUPS_FILE_NAME'])
    parser.add_argument('-m', '--moodle-students-list', required=True,
                        help='The list of all students with their group memberships as exported by moodle.', type=str)
    parser.add_argument('-gfmts', '--group-formats', required=False,
                        help='The format strings of the group names used in grouping creation. For example, "Group A1 {:02}" which corresponds to names with 2-padded group ids, e.g., "Group A1 01".',
                        type=str, nargs='+', default=defaults['DEFAULT_GROUP_FORMATS'])
    parser.add_argument('-gns', '--grouping-names', required=False,
                        help='The names of the groupings. For example, "Assignment Part 1" "Assignment Part 2".',
                        type=str, nargs='+', default=defaults['DEFAULT_GROUPING_NAMES'])
    parser.add_argument('-v', '--verbose', required=False, default=False, action='store_true')


class GenMiscTable(CommandModule):
    module_name = 'gen-misc'
    commands = [('groups-info', register_gen_group_info, gen_group_info_table)]

    @property
    def module_properties(self) -> dict[str, Any]:
        return {
            'force_subparsers': True
        }

    @property
    def additional_config(self) -> dict[str, Any]:
        nd = self.default_config['naming_dictionary']
        return {
            'DEFAULT_GROUP_FORMATS': ['Group A1 {:03}', 'Group A2 {:03}'],
            'DEFAULT_GROUPS_FILE_NAME': 'moodle-groups-info.xlsx',
            'DEFAULT_GROUPING_NAMES': ['Assignment Part 1', 'Assignment Part 2'],
        }


if __name__ == '__main__':
    GenMiscTable().as_program('gen').parse_and_run()



import argparse
import itertools
import os.path
from typing import Any

import pandas as pd

from grading_tools.common.commands import CommandModule
from grading_tools.common.defaults import NamingDictionary
from grading_tools.common.gradable_spec import load_spec
from grading_tools.common.utils import read_generic_table
from grading_tools.grading_table_gen.gen_grading_table import save_grading_table, \
    register_gen_grading_table_options
from grading_tools.grading_table_gen.grading_table_commons import mk_node_name


def read_stripped_df(file_name: str, index_columns: list[str], question_columns: list[str],
                     naming_dictionary: NamingDictionary) -> pd.DataFrame:
    df = read_generic_table(file_name, table_name=naming_dictionary.GRADING_TABLE_NAME)
    filtered_columns = [c for c in df.columns if c in set(question_columns)]
    df = df.set_index(index_columns, drop=True)
    df = df[filtered_columns]  # .astype('Float64')
    stripped = df.dropna(axis='columns', how='all').dropna(axis='rows', how='all')
    return stripped


def read_merge_pattern(merge_pattern: str) -> list[list[int]]:
    pattern = [[int(s.strip()) for s in row.split(' ')] for row in merge_pattern.split('/')]
    assert all(len(row) > 0 for row in pattern)
    return pattern


def merge_dfs(df_table: list[list[pd.DataFrame]]) -> pd.DataFrame:
    merged_row_dfs = []
    for row in df_table:
        merged_row_df = pd.concat(row, axis='columns', join='outer', verify_integrity=True)
        merged_row_dfs.append(merged_row_df)
    merged_df = pd.concat(merged_row_dfs, axis='index', verify_integrity=True)
    return merged_df


def assert_non_overlapping(df_table: list[list[pd.DataFrame]]):
    idx_sets = []
    for row in df_table:
        sets = [set(df.columns) for df in row]
        # mutually disjoint
        assert all(set.isdisjoint(a, b) for a, b in itertools.combinations(sets,
                                                                           2)), 'The columns of the individual grading table columns within a merged row must be disjoint. This is to ensure no values are accidentally overwritten.'
        idx_sets.append(set.union(*(set(df.index) for df in row)))
    assert all(set.isdisjoint(a, b) for a, b in itertools.combinations(idx_sets,
                                                                       2)), 'The indices of the individual grading table rows must be disjoint. This is to ensure no values are accidentally overwritten.'


def apply_pattern(dfs: list[pd.DataFrame], pattern: list[list[int]]) -> list[list[pd.DataFrame]]:
    return [[dfs[i] for i in row] for row in pattern]


def merge_grading(merge_pattern: str = None, files: list[str] = None, spec_file: str = None,
                  index_columns: list[str] = None,
                  naming_dictionary: NamingDictionary = None, **config):
    spec = load_spec(spec_file)

    # the condition skips any skippable, e.g., total/summary nodes
    question_columns = [mk_node_name(n) for n in spec.tree.leaves_iter() if not n.get_property('skip')]

    pattern = read_merge_pattern(merge_pattern)
    dfs = [read_stripped_df(f, index_columns, question_columns, naming_dictionary) for f in files]
    df_table = apply_pattern(dfs, pattern)

    assert_non_overlapping(df_table)
    merged_df = merge_dfs(df_table)

    if config['verbose']:
        print('Merged DF')
        print(merged_df)

    save_grading_table(spec, data_df=merged_df, output_sheet=naming_dictionary.GRADING_SHEET_NAME,
                       output_table=naming_dictionary.GRADING_TABLE_NAME, naming_dictionary=naming_dictionary, **config)


def update_grading(spec_file: str = None, grading_file: str = None, **config):
    if not config.get('output_file'):
        config['output_file'] = os.path.splitext(grading_file)[0] + '-updated.xlsx'
    merge_grading(merge_pattern='0', spec_file=spec_file, files=[grading_file], **config)


MERGE_PATTERN_REGEX = r'(\d+)(\s+(\d+))*(\/(\d+)(\s+(\d+))*)*'


def register_parser_merge(parser: argparse.ArgumentParser, **defaults):
    parser.add_argument('-o', '--output', required=False, help='Filename of the output.', dest='output_file')
    parser.add_argument('-s', '--spec-file', type=str, required=True,
                        help='Path to the specification of the resulting gradable.')
    parser.add_argument('-f', '--files', type=str, nargs='+', required=True, help='Paths to the files to merge.')
    parser.add_argument('-mp', '--merge-pattern', required=True,
                        help=f'String specifying the merge pattern. Format: "{MERGE_PATTERN_REGEX}", e.g., "1 0 2/3" will merge the files with indices 1, 0 and 2 horizontally, i.e., join them, and then merge file 3 onto the result vertically, i.e., by row concatenation.',
                        default=defaults['DEFAULT_MERGE_PATTERN'])

    parser.add_argument('-ic', '--index-columns', nargs='*', required=False,
                        help='The column names to use as the index. Used for joining, as well as creation of the resulting grading table.',
                        default=defaults['DEFAULT_INDEX_COLUMNS'])

    register_gen_grading_table_options(parser, **defaults)


def register_parser_update(parser: argparse.ArgumentParser, **defaults):
    parser.add_argument('-o', '--output', required=False, help='Filename of the output.', dest='output_file')
    parser.add_argument('-s', '--spec-file', type=str, required=True,
                        help='Path to the updated specification of this gradable.')
    parser.add_argument('-g', '--grading-file', required=True, help='Path to the grading file to update.')

    register_gen_grading_table_options(parser, **defaults)


class EditGradingTable(CommandModule):
    module_name = 'edit-grading'
    commands = [('merge', register_parser_merge, merge_grading), ('update', register_parser_update, update_grading)]

    @property
    def additional_config(self) -> dict[str, Any]:
        nd = self.default_config['naming_dictionary']
        return {'DEFAULT_MERGE_PATTERN': '0', 'DEFAULT_INDEX_COLUMNS': list(nd.IDENTITY_COLS) + [nd.MATR_COL]} | {
            'DEFAULT_OUTPUT_FILENAME': ''}


if __name__ == '__main__':
    EditGradingTable().as_program('edit-grading').parse_and_run()



from __future__ import annotations

import argparse
import itertools
from typing import Literal

import pandas as pd

from grading_tools.common.commands import CommandModule
from grading_tools.common.defaults import NamingDictionary
from grading_tools.common.excel_utils import write_to_excel
from grading_tools.common.gradable_spec import *
from grading_tools.common.utils import complete_file_path
from grading_tools.grading_table_gen.grading_table_commons import mk_node_name, \
    mk_sum_formula_range, mk_header_info, mk_initialization_df, \
    mk_marked_formula_over_ranges, mk_sum_formula_over_ranges, mk_mandatory_col_condition, \
    mk_is_checked_condition, COLUMN_NAME_LEVEL_START

GradingStyle = Literal['numbers', 'checkboxes']


def merge_length(path: tuple[str, ...]) -> int | None:
    last_value = path[-1]
    last_index = None
    for i, n in enumerate(reversed(path[:-1])):
        if n != last_value:
            last_index = i
            break
    if last_index is None:
        return len(path)
    elif last_index > 0:
        return last_index
    else:
        return None


def add_visual_table_header(exwr: pd.ExcelWriter, output_sheet: str, output_table: str, startrow: int, startcol: int, *,
                            spec: GradableSpecV1, header_index: pd.Index, paths_list: list[tuple[str, ...]],
                            naming_dictionary: NamingDictionary, **config):
    header_height = header_index.nlevels
    pd.DataFrame(columns=header_index).to_excel(exwr, sheet_name=output_sheet, startrow=startrow, startcol=startcol)
    sheet = exwr.sheets[output_sheet]
    bold_merge_format = exwr.book.add_format({'align': 'center', 'bold': True, 'border': 1})
    merge_format = exwr.book.add_format({'align': 'center'})
    merge_format.set_align('vcenter')

    for i, p in enumerate(paths_list, start=1):
        if (j := merge_length(p)) is not None:
            sheet.merge_range(startrow + len(p) - 1 - j, startcol + i, startrow + len(p) - 1, startcol + i, p[-1],
                              bold_merge_format)
    if startcol > 0:
        for i, s in enumerate(header_index.names):
            sheet.merge_range(startrow + i, 0, startrow + i, startcol, s, merge_format)
    return header_height


def add_inline_aux_info(exwr: pd.ExcelWriter, output_sheet: str, output_table: str, startrow: int, startcol: int, *,
                        grading_style: GradingStyle, spec: GradableSpecV1, column_names_list: list[str],
                        pts_list: list[float], naming_dictionary: NamingDictionary, **config) -> int:
    if grading_style == 'numbers':
        pd.DataFrame(data=[pts_list], dtype='Float64', index=['Max Points']).to_excel(exwr,
                                                                                      startrow=startrow,
                                                                                      startcol=startcol,
                                                                                      sheet_name=output_sheet,
                                                                                      index=True,
                                                                                      header=False)
        return 1
    elif grading_style == 'checkboxes':
        book = exwr.book
        sheet = exwr.sheets[output_sheet]
        bold_format = book.add_format({'bold': True})
        bold_extra_format = book.add_format({'bold': True, 'bottom': 6})
        vert_format = book.add_format({'rotation': 90})
        merge_format = book.add_format({'align': 'center'})

        pd.DataFrame(data=[pts_list], dtype='Float64', index=['Max Points']).to_excel(exwr,
                                                                                      startrow=startrow + 1,
                                                                                      startcol=startcol + 1,
                                                                                      sheet_name=output_sheet,
                                                                                      index=False,
                                                                                      header=False)
        sheet.write(startrow + 1, startcol, 'Points')
        sheet.write(startrow + 2, startcol, 'Stats')
        sheet.write(startrow + 3, startcol, 'Text')
        for i, (c, n) in enumerate(zip(column_names_list, spec.tree.leaves_iter()), start=1):
            if not n.get_property('skip') and not n.get_property('skip_in_formula'):
                rg, crit = mk_is_checked_condition(c, config["CHECKBOX_SYMBOL"], table=output_table)
                sheet.write(startrow + 2, startcol + i,
                            f'=COUNTIF({rg},{crit})',
                            bold_format)
            elif n.get_property('summary'):
                sheet.write(startrow + 2, startcol + i,
                            f'=IF(COUNT({output_table}[{c}])>0,AVERAGE({output_table}[{c}]),"")', bold_extra_format)

            text = n.get_property('text')
            if not text:
                text = ''
            sheet.write(startrow + 3, startcol + i, text, vert_format)
        if startcol > 0:
            sheet.merge_range(startrow, 0, startrow + 3, startcol - 1, '', merge_format)

        excel_cols = [{'header': 'Info'}] + [{'header': c} for c in column_names_list]
        extra_options = {}
        sheet.add_table(startrow, startcol, startrow + 3, startcol + len(excel_cols) - 1,
                        {'name': naming_dictionary.GRADING_GKINFO_TABLE_NAME, 'columns': excel_cols} | extra_options)
        return 4
    else:
        return 0


def add_grading_table(exwr: pd.ExcelWriter, output_sheet: str, output_table: str, startrow: int, startcol: int, *,
                      spec: GradableSpecV1,
                      initialization_df: pd.DataFrame,
                      grading_style: GradingStyle = 'numbers',
                      utility_columns: bool | None = False,
                      cap_bonus_pts: bool = False,
                      naming_dictionary: NamingDictionary = None, **config):
    w_book = exwr.book
    regular_format = w_book.add_format({'align': 'center'})
    summary_format = w_book.add_format({'bold': True, 'align': 'center'})

    green_format = w_book.add_format({'bg_color': '#7df59f'})
    red_format = w_book.add_format({'bg_color': '#FFC7CE', 'font_color': '#9C0006'})
    w_sheet_scores = exwr.sheets[output_sheet]

    initialization_df.to_excel(exwr, sheet_name=output_sheet, float_format='%.2f', index=True,
                               header=False,
                               startrow=startrow + 1, startcol=startcol, merge_cells=False)
    # pd.DataFrame(initialization_df.values).to_excel(exwr, sheet_name=output_sheet, float_format='%.2f', index=False,
    #                                          header=False,
    #                                          startcol=startcol + 1, startrow=current_row, merge_cells=False)
    # initialization_df.index.to_excel(exwr, sheet_name=output_sheet, index=False, header=False,
    #                                         startrow=current_row, startcol=0)

    (table_rows, table_cols) = initialization_df.shape
    # w_sheet_scores.set_column(1, max_col, 10)
    table_start = startrow
    sum_formula_func = mk_sum_formula_over_ranges if grading_style == 'numbers' else mk_marked_formula_over_ranges

    array_formulas = []
    bound_conditions = []
    equality_conditions = []
    excel_cols = [{'header': name if name else f'IndexCol-{i}'} for i, name in enumerate(initialization_df.index.names)]
    extra_options = {}
    for column_index, leaf in enumerate(spec.tree.leaves_iter(), start=len(excel_cols)):

        col = {'header': mk_node_name(leaf), 'format': regular_format}
        if leaf.get_property('summary'):
            summary_range = itertools.chain(
                *(c.leaves_iter() for c in leaf.left_siblings))

            local_ranges, local_max_pts = mk_sum_formula_range(summary_range)

            total_formula, is_arr_formula = sum_formula_func(local_ranges, output_table,
                                                             max_points=(local_max_pts if cap_bonus_pts else None),
                                                             naming_dictionary=naming_dictionary,
                                                             **config)
            if is_arr_formula:
                array_formulas.append((column_index, total_formula))
            else:
                col.update({'formula': total_formula})
            col.update({'format': summary_format})
            if grading_style == 'checkboxes':
                lb, ub = (leaf.pts, 0) if leaf.pts < 0 else (0, leaf.pts)
                bound_conditions.append((column_index, False, lb, ub))
        else:
            if grading_style == 'numbers':
                lb, ub = (leaf.pts, 0) if leaf.pts < 0 else (0, leaf.pts)
                bound_conditions.append((column_index, False, lb, ub))
        if leaf.get_property('mandatory'):
            equality_conditions.append((column_index, True, f'"{config["CHECKBOX_SYMBOL"]}"'))

        excel_cols.append(col)

    if utility_columns:
        all_ranges, total_max_pts = mk_sum_formula_range(spec.tree.leaves_iter())

        participation_formula = ''
        if grading_style == 'numbers':
            rng = f'[@[{excel_cols[0]["header"]}]:[{excel_cols[-1]["header"]}]]'
            participation_formula = f'COUNTA({rng})=COLUMNS({rng})'
        elif grading_style == 'checkboxes':
            id_col_rng = f'[@[{initialization_df.index.names[0]}]:[{initialization_df.index.names[-1]}]]'
            mandatory_cols = mk_mandatory_col_condition(all_ranges, **config)
            identity_cols = f'COUNTA({id_col_rng})=COLUMNS({id_col_rng})'
            participation_formula = f'AND({identity_cols}{("," + mandatory_cols) if mandatory_cols else ""})'
        excel_cols.append(
            {'header': naming_dictionary.PARTICIPATED_COL, 'formula': participation_formula})

        total_formula, is_arr_formula = sum_formula_func(all_ranges, output_table,
                                                         max_points=(total_max_pts if cap_bonus_pts else None),
                                                         with_equals=False,
                                                         naming_dictionary=naming_dictionary,
                                                         **config)
        total_formula = f'IF([@[{naming_dictionary.PARTICIPATED_COL}]], {total_formula}, "{config["FALLBACK_SCORE"]}")'
        total_col = {'header': naming_dictionary.GRADING_TOTAL_COL}
        if is_arr_formula:
            array_formulas.append((len(excel_cols), total_formula))
        else:
            total_col.update({'formula': total_formula})
        excel_cols.append(total_col)
        extra_options = {'last_column': True}

    w_sheet_scores.add_table(table_start, startcol, table_start + table_rows, startcol + len(excel_cols) - 1,
                             {'name': output_table,
                              'columns': excel_cols} | extra_options)
    for i, formula in array_formulas:
        w_sheet_scores.write_array_formula(table_start + 1, startcol + i, table_start + table_rows, startcol + i,
                                           '{' + formula + '}')

    if config['conditional_formatting']:
        for i, is_pos, lb, ub in bound_conditions:
            # conditional formatting is somehow broken
            w_sheet_scores.conditional_format(startrow + 1, startcol + i, startrow + table_rows, startcol + i,
                                              {'type': 'cell', 'criteria': 'between' if is_pos else 'not between',
                                               'minimum': lb,
                                               'maximum': ub, 'format': green_format if is_pos else red_format})
        for i, is_pos, equals in equality_conditions:
            condition = {'type': 'text', 'criteria': 'containing' if is_pos else 'not containing',
                         'value': equals[1:-1], 'format': green_format if is_pos else red_format}
            w_sheet_scores.conditional_format(startrow + 1, startcol + i, startrow + table_rows, startcol + i,
                                              condition)

    return len(initialization_df) + 1


def add_aux_sheet(exwr: pd.ExcelWriter, header_index: pd.Index, pts_list: list[float], **config):
    ## aux. sheet
    aux_sheet_name = config['AUX_SHEET_NAME']
    pts_df = pd.DataFrame(data={'Max Points': pts_list}, index=header_index, dtype='Float64')
    pts_df.to_excel(exwr, sheet_name=aux_sheet_name, float_format='%.2f')
    w_sheet_aux = exwr.sheets[aux_sheet_name]
    for i in range(header_index.nlevels + 1):
        w_sheet_aux.set_column(0, 1 + i, 15)


def save_grading_table(spec: GradableSpecV1,
                       output_sheet: str = None,
                       output_table: str = None,
                       verbose: bool = False,
                       cutoff_level: int | str | None = None,
                       totals_level: int | str | None = None,
                       **config):
    if verbose:
        print(spec)
        print(config)

    header_index, column_names_list, pts_list, paths_list = mk_header_info(spec, cutoff_level, totals_level, **config)

    initialization_df = mk_initialization_df(spec, column_names_list, **config)

    extra_args = dict(spec=spec, header_index=header_index, column_names_list=column_names_list, pts_list=pts_list,
                      paths_list=paths_list,
                      initialization_df=initialization_df)  # , naming_dictionary=config.get('naming_dictionary')

    def my_write(exwr: pd.ExcelWriter):
        current_row = 0
        offset_startcol = initialization_df.index.nlevels - 1

        current_row += add_visual_table_header(exwr, output_sheet, output_table, current_row, offset_startcol,
                                               **extra_args, **config)

        # auxiliary inline table/info
        # has to return the number of rows written
        current_row += add_inline_aux_info(exwr, output_sheet, output_table, current_row, offset_startcol, **extra_args,
                                           **config)

        current_row += add_grading_table(exwr, output_sheet, output_table, current_row, 0, **extra_args, **config)

        if config['aux_sheet']:
            add_aux_sheet(exwr, **extra_args, **config)

    write_to_excel(
        complete_file_path(config['output_file'], default_file_name=config.pop('DEFAULT_GRADING_FILE_NAME', None)),
        my_write, err_on_file_existence=False)


def gen_grading_table(spec_file: str, cap_bonus_pts=False, **config):
    spec = load_spec(spec_file)

    root = spec.tree
    cap = sum(n.pts for n in root.leaves_iter() if
              not n.get_property('bonus') and not n.get_property('skip') and not n.get_property('skip_in_total'))
    bonus = sum(n.pts for n in root.leaves_iter() if n.get_property('bonus'))
    print(f'Points: {cap} + {bonus} (bonus) = {root.pts}')
    if cap_bonus_pts and bonus > 0:
        print(f'Points formula is capped to {cap} (MIN({cap},SUM(...))).')

    save_grading_table(spec, **config)


def register_gen_grading_table_options(parser: argparse.ArgumentParser, **defaults):
    parser.add_argument('-gs', '--grading-style', required=False, choices=['numbers', 'checkboxes'], default='numbers',
                        help='The grading style to use: "numbers" or "checkboxes". Note that these require different specification file properties.')

    parser.add_argument('-cl', '--cutoff-level', required=False, default=None, type=str,
                        help='Level up to which to include (sub^x) questions. E.g. choosing "Question" cuts off/aggregates subquestions.')
    parser.add_argument('-tl', '--totals-level', required=False, default=None, type=str,
                        help='Level up to which to introduce (sub^x) question totals columns. E.g. choosing "Question" introduces per-question totals.')

    parser.add_argument('-uc', '--utility-columns', required=False, action='store_false', default=True,
                        help='Whether to include utility columns.')
    parser.add_argument('-cf', '--conditional-formatting', required=False, action='store_false', default=True,
                        help='Whether to include add conditional formatting for some sanity checks.')
    parser.add_argument('-cap', '--cap-bonus-pts', required=False, action='store_true', default=False,
                        help='Whether to cap total pts to maximum without bonus.')
    parser.add_argument('-v', '--verbose', action='store_true', required=False, default=False,
                        help='Whether to print verbose output.')
    parser.add_argument('-aux', '--aux-sheet', required=False, action='store_true', default=False,
                        help='Whether to include an auxiliary sheet with reformatted information.')


def register_gen_grading_table(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults):
    parser.add_argument('-s', '--spec-file', required=True, type=str, help='Path to spec file.')

    parser.add_argument('-o', '--output-file', required=False, type=str, help='Path to output file.',
                        default=defaults['DEFAULT_GRADING_FILE_NAME'])
    parser.add_argument('-ot', '--output-table', required=False, type=str, help='Name of the table to generate.',
                        default=defaults['DEFAULT_TABLE_NAME'])
    parser.add_argument('-osh', '--output-sheet', required=False, type=str,
                        help='Sheet where to put the generated table.',
                        default=defaults['DEFAULT_SHEET_NAME'])

    parser.add_argument('-df', '--data-file', required=False, type=str,
                        help='Path to (bootstrapping) data file. This file is used to prefill the grading table. If this option is used, the index information is ignored.')
    parser.add_argument('-dsh', '--data-sheet', required=False, type=str,
                        help='Optionally, an excel sheet that contains the data to use.',
                        default=naming_dictionary.GRADING_SHEET_NAME)
    parser.add_argument('-dt', '--data-table', required=False, type=str,
                        help='Optionally, an excel table that contains the data to use.',
                        default=naming_dictionary.GRADING_TABLE_NAME)

    parser.add_argument('-if', '--index-file', required=False, type=str,
                        help='Optionally, a file that contains an index to use. For example, a list of matriculation numbers.')
    parser.add_argument('-ift', '--index-file-preset', required=False, choices=('manual', 'rwthonline'), type=str,
                        default='manual',
                        help='Optionally, specify the a predefined index file handling to enable specific parsing, e.g., of RWTHOnline exam registrations. Overrides the other index options.')
    parser.add_argument('-ish', '--index-sheet', required=False, type=str,
                        help='Optionally, an excel sheet that contains the index to use.')
    parser.add_argument('-it', '--index-table', required=False, type=str,
                        help='Optionally, an excel table that contains the index to use.')
    parser.add_argument('-ic', '--index-columns', nargs='*', required=False,
                        help='Optionally, the column names of the index file to use as the index. If no file is provided, only the names are used.',
                        default=defaults['DEFAULT_INDEX_COLUMNS'])

    register_gen_grading_table_options(parser, **defaults)


class GenGradingTable(CommandModule):
    module_name = 'gen-grading'
    commands = [('grading', register_gen_grading_table, gen_grading_table)]

    @property
    def additional_config(self) -> dict[str, Any]:
        nd = self.default_config['naming_dictionary']
        return {
            'DEFAULT_GRADING_FILE_NAME': 'grading-empty.xlsx',
            'DEFAULT_INDEX_COLUMNS': list(nd.IDENTITY_COLS) + [nd.MATR_COL],
            'AUX_SHEET_NAME': 'Aux',
            'DEFAULT_SHEET_NAME': nd.GRADING_SHEET_NAME,
            'DEFAULT_TABLE_NAME': nd.GRADING_TABLE_NAME,
            'FALLBACK_SCORE': '',
            'CHECKBOX_SYMBOL': 'x',
            'placeholder_row_count': 10,
            'upper_label_level': COLUMN_NAME_LEVEL_START,
        }


if __name__ == '__main__':
    GenGradingTable().as_program('gen').parse_and_run()



import argparse
import glob
import itertools
import os
import pathlib
from collections import defaultdict
from typing import Any
import yaml
import jsonmerge

from grading_tools.common.commands import CommandModule
from grading_tools.common.gradable_spec import GradableSpecV1


def mk_merge_schema(depth: int):
    schema = {"type": "object",
              "properties": {
                  "tree-level-names": {
                      "items": {
                          "mergeStrategy": "override"
                      }
                  },
                  "tree": {
                      "properties": {}
                  }}}
    rec = schema["properties"]["tree"]
    for _ in range(depth):
        rec["properties"] = {}
        rec["properties"]["children"] = {
            "mergeStrategy": "append",  # Recursively merge nested objects
            "items": {

            }
        }
        rec = rec["properties"]["children"]["items"]
    return schema


def concat_specs(spec_files: list[str] = None, output_file: str = None, max_depth: int = None, **config):
    specs = []
    for spec_file in spec_files:
        with open(spec_file, 'r') as f:
            specs.append(yaml.load(f, yaml.FullLoader))

    merger = jsonmerge.Merger(mk_merge_schema(max_depth))
    result = merger.merge(*specs)

    if (dir_name := os.path.dirname(output_file)) != '':
        os.makedirs(dir_name, exist_ok=True)
    with open(output_file, 'w', encoding='utf-8') as f:
        yaml.dump(dict(result), f, sort_keys=False)


def register_parser_concat(parser: argparse.ArgumentParser, **defaults):
    parser.add_argument('-o', '--output', required=False, help='Filename of the output.', dest='output_file',
                        default=defaults['DEFAULT_OUTPUT_FILE'])
    parser.add_argument('-s', '--spec-files', type=str, required=True, nargs='+',
                        help='Path to the specs to concatenate.')
    parser.add_argument('-md', '--max-depth', type=int, required=False, help='Maximum depth to merge nodes at',
                        default=defaults['DEFAULT_MAX_DEPTH'])


class ConcatSpecs(CommandModule):
    module_name = 'concat-specs'
    commands = [('concat', register_parser_concat, concat_specs)]

    @property
    def additional_config(self) -> dict[str, Any]:
        nd = self.default_config['naming_dictionary']

        return {
            'DEFAULT_OUTPUT_FILE': 'specs-concat.yaml',
            'DEFAULT_TREE_LEVEL_NAMES': ['Assignment', 'Question', 'Subquestion', 'Grading Key'],
            'DEFAULT_MAX_DEPTH': 6
        }


if __name__ == '__main__':
    ConcatSpecs().as_program('gen').parse_and_run()



import argparse
import itertools
import typing
from collections.abc import Iterable
from typing import Sequence

import pydantic
import yaml

from grading_tools import mk_cli_program
from grading_tools.common.commands import CommandModule, CliProgram
from grading_tools.common.defaults import NamingDictionary, get_base_cfg

OptionsList = list[str]


class Flags(typing.TypedDict, total=False):
    options: OptionsList | None


CommandArgs = Flags
ModuleArgs = dict[str, CommandArgs]


class FileCFG(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(extra='allow')

    naming_dictionary: NamingDictionary = pydantic.Field(default_factory=NamingDictionary)
    arguments: CommandArgs | ModuleArgs = pydantic.Field(default_factory=dict)


def dig_dict(dic: dict, keys: Iterable[str]) -> dict:
    res = dic
    i = 0
    for k in keys:
        if k in res:
            i += 1
        res = res.get(k, {})
    return res


def expand_flag(flag: tuple[str, str | list[str]]) -> tuple[str, ...]:
    k, v = flag
    if v is None or v == 'null':
        return (k,)
    if isinstance(v, str):
        return k, v
    elif isinstance(v, list):
        return (k,) + tuple(v)
    else:
        return k, str(v)


def call(config_file: str = None, command: list[str] = None, path: list[list[str]] = None, remainder: list[str] = None,
         **config):
    cfg: FileCFG
    with open(config_file, 'r') as f:
        cfg = FileCFG.model_validate(yaml.full_load(f))

    if len(path) == 0:
        path = [[]]

    configured_arguments = {}
    for path_ in path:
        specific_args = dig_dict(cfg.arguments, itertools.chain(command, path_))

        if not specific_args:
            print(f'There is no config for {command} under path {path_} in the cfg file: {config_file}.')

        configured_arguments.update(specific_args)

    arg_list = list(command)

    if 'options' in configured_arguments:
        options = configured_arguments.pop('options')
        arg_list.extend(options)

    arg_list.extend(itertools.chain(*(expand_flag(flag) for flag in configured_arguments.items())))

    if remainder:
        arg_list.extend(remainder)

    merged_naming_dictionary = NamingDictionary(
        **get_base_cfg().get('naming_dictionary', {}).model_dump(exclude_defaults=True, exclude_unset=True))

    cli_program = mk_cli_program(default_config={'naming_dictionary': merged_naming_dictionary})

    cli_program.parse_and_run(arg_list)


def register_parser(parser: argparse.ArgumentParser, **defaults):
    parser.add_argument('-cfg', '--config-file', type=str, help='Path to the configuration file.', required=True)
    parser.add_argument('-cmd', '--command', type=str, help='The command to execute.', required=True, nargs='+')
    parser.add_argument('-p', '--path', type=str,
                        help='Optionally, a path to the specific config for the command. If this argument is used multiple times, the configs are applied in order, i.e., potentially overriding the previous ones. This allows creating shared base configs.',
                        required=False, nargs='+', default=[], action='append')
    # + 'remainder' which is set explicitly


class FromCfg(CommandModule):
    module_name = 'from-cfg'
    commands = [('from-cfg', register_parser, call)]


class CfgProgram(CliProgram):

    def parse_and_run(self, args: Sequence[str] | None = None) -> None:
        args, extra = self.make_parser().parse_known_args(args)
        args.remainder = extra
        args.func(args)


def mk_cfg_cli_program():
    return CfgProgram('from-cfg', FromCfg, *mk_cli_program().module_classes)


if __name__ == '__main__':
    mk_cfg_cli_program().parse_and_run()



import argparse
import os
import zipfile
from typing import Any

import pandas as pd

from grading_tools.common.commands import CommandModule
from grading_tools.common.compat_utils import stringify_scores_for_moodle
from grading_tools.common.defaults import NamingDictionary
from grading_tools.common.utils import read_generic_table


def generate_feedback(grading_file: str, gk_info_file: str, naming_dictionary: NamingDictionary, **config) -> tuple[
    dict[int, str], pd.DataFrame]:
    grading_df = read_generic_table(grading_file, sheet_name=config.pop('grading_sheet', None),
                                    table_name=config.pop('grading_table', None))
    gk_info_df = read_generic_table(gk_info_file, sheet_name=config.pop('gk_info_sheet', None),
                                    table_name=config.pop('gk_info_table', None))
    gk_info_df = gk_info_df.set_index('Info')
    gk_info_df_t = gk_info_df.T.convert_dtypes(convert_string=True, convert_floating=True)

    grading_df = grading_df.set_index(naming_dictionary.GRADING_GROUP_ID_COL)
    gk_text = gk_info_df_t['Text']
    gk_pts = gk_info_df_t['Points']

    bitmask = grading_df[gk_info_df_t.index].map(lambda x: x and type(x) is str and ('x' in x or 'X' in x)).astype(bool)

    if config.get('verbose', False):
        print(gk_info_df)
        print(grading_df)
        print(bitmask)
        print(bitmask.sum(axis=1))

    feedbacks = {}
    for group_id in bitmask.index:
        row = bitmask.loc[group_id].to_numpy()
        qs = gk_info_df_t.index[row]
        strs = gk_text.iloc[row]
        pts = gk_pts.iloc[row]

        def fmt_q(q: str):
            return '-'.join(q.split('-')[:-1])

        def fmt_t(t: str | None):
            return t + ' ' if pd.notna(t) else ''

        txt = f'Feedback Group {group_id}\n'
        txt += '\n'.join((f'{fmt_q(q)}: {fmt_t(t)}({p})' for q, t, p in zip(qs, strs, pts)))
        feedbacks[group_id] = txt

    data = {'total': stringify_scores_for_moodle(grading_df[naming_dictionary.GRADING_TOTAL_COL], naming_dictionary),
            'feedback': [feedbacks[gid] for gid in grading_df.index]}  # feedbacks.values() should also work
    feedback_df = pd.DataFrame(data, index=pd.Index(grading_df.index, name=naming_dictionary.GRADING_GROUP_ID_COL))

    return feedbacks, feedback_df


def zip_round_trip(feedbacks: dict[int, str], output_zip: str, submissions_file: str, group_name_fmt: str, **config):
    dirname = os.path.dirname(output_zip)
    if dirname:
        os.makedirs(dirname, exist_ok=True)
    with zipfile.ZipFile(submissions_file, 'r', zipfile.ZIP_DEFLATED) as sz:
        submission_names = set(sz.namelist())

        with zipfile.ZipFile(output_zip, 'w', zipfile.ZIP_DEFLATED) as fz:
            for group_id, txt in feedbacks.items():
                group_name = group_name_fmt.format(group_id)
                matching_submissions = {sn for sn in submission_names if group_name in sn}
                if config.get('verbose', False):
                    if len(matching_submissions) > 1:
                        print(f'Multiple submission folders match the group name {group_name} for id {group_id}.')
                    if len(matching_submissions) < 1:
                        print(f'No submission folders match the group name {group_name} for id {group_id}.')
                if len(matching_submissions) == 1:
                    submission_name = matching_submissions.pop()
                    fz.writestr(os.path.join(submission_name, 'feedback.txt'), txt)


def export_feedback_folder(output_dir: str, grading_file: str, gk_info_file: str, naming_dictionary: NamingDictionary,
                           **config):
    feedbacks, feedback_df = generate_feedback(grading_file, gk_info_file, naming_dictionary, **config)

    os.makedirs(output_dir, exist_ok=True)
    for group_id, txt in feedbacks.items():
        with open(os.path.join(output_dir, f'feedback-{group_id}.txt'), 'w') as f:
            f.write(txt)
    feedback_df.to_csv(os.path.join(output_dir, 'all-feedback.csv'))


def export_feedback_zip(moodle_output: str, submissions_zip: str, grading_file: str, gk_info_file: str,
                        naming_dictionary: NamingDictionary, **config):
    feedbacks, feedback_df = generate_feedback(grading_file, gk_info_file, naming_dictionary, **config)
    zip_round_trip(feedbacks, moodle_output, submissions_zip, **config)


def export_feedback(**config):
    if 'output_dir' in config:
        export_feedback_folder(**config)
    if 'moodle_output' in config:
        export_feedback_zip(**config)


def configure_parser(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults):
    parser.add_argument('-o', '--output', required=False, help='Path to output directory.', dest='output_dir',
                        default=defaults['DEFAULT_OUTPUT_DIR_NAME'])

    parser.add_argument('-g', '--grading-file', required=True, help='Path to grading file with group scores.')
    parser.add_argument('-gk', '--gk-info-file', required=True, help='Path to grading key info.',
                        dest='gk_info_file')

    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-sh', '--grading-sheet', required=False, help='Optionally, excel sheet within grading file.')
    group.add_argument('-t', '--grading-table', required=False, help='Optionally, excel table within grading file.',
                       default=naming_dictionary.GRADING_TABLE_NAME)

    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-gksh', '--gk-info-sheet', required=False,
                       help='Optionally, gk info excel sheet within grading file.')
    group.add_argument('-gkt', '--gk-info-table', required=False,
                       help='Optionally, gk info excel table within grading file.',
                       default=naming_dictionary.GRADING_GKINFO_TABLE_NAME)

    parser.add_argument('-v', '--verbose', required=False, action='store_true', default=False,
                        help='Whether to produce more verbose output.')


def configure_parser_moodle(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults):
    configure_parser(parser, naming_dictionary, **defaults)
    parser.add_argument('-mo', '--moodle-output', required=False, help='Path to moodle output zip.',
                        default=defaults['DEFAULT_MOODLE_OUTPUT_NAME'])

    parser.add_argument('-szip', '--submissions-zip', required=True,
                        help='Path to submissions zip file as downloaded from moodle.')
    parser.add_argument('-gfmt', '--group-name-fmt', required=True,
                        help='Format str to recreate the group names used in moodle group creation (in Python style, e.g., "Group A1 {:02}").')


class ComposeFeedback(CommandModule):
    module_name = 'gen-group-feedback'
    commands = [('basic', configure_parser, export_feedback_folder),
                ('moodle', configure_parser_moodle, export_feedback_zip)]

    @property
    def additional_config(self) -> dict[str, Any]:
        return {'DEFAULT_OUTPUT_DIR_NAME': 'feedback/', 'DEFAULT_MOODLE_OUTPUT_NAME': 'moodle-feedback.zip'}


if __name__ == '__main__':
    ComposeFeedback().as_program('group-feedback').parse_and_run()



import argparse
from datetime import datetime
from pathlib import Path
from typing import Any

import pandas as pd

from grading_tools.common.commands import CommandModule
from grading_tools.common.defaults import *
from grading_tools.common.excel_utils import write_table_to_excel
from grading_tools.common.compat_utils import stringify_scores_for_moodle
from grading_tools.common.utils import read_moodle_csv, write_moodle_csv, read_generic_table


def print_rows_pretty(df, new_scores, naming_dictionary: NamingDictionary):
    print_df = pd.DataFrame(df[[naming_dictionary.MOODLE_MATR_COL] + list(naming_dictionary.MOODLE_IDENTITY_COLS)])
    print_df['OLD score'] = df[naming_dictionary.MOODLE_GRADING_COL]
    print_df['NEW score'] = new_scores
    print(print_df.reset_index(drop=True))


# misc


def join_moodle_and_scores(grading_file: str, moodle_grading_worksheet: str, group_name_fmt: str,
                           naming_dictionary: NamingDictionary, **config) -> tuple[pd.DataFrame, list[str]]:
    if config['verbose']:
        print(config)
    grading_column = config['grading_column']
    feedback_column = config['feedback_column']
    grading_df = read_generic_table(grading_file, sheet_name=config.pop('grading_sheet', None),
                                    table_name=config.pop('grading_table', None))
    moodle_df = read_moodle_csv(moodle_grading_worksheet)
    original_cols = moodle_df.columns

    # group id (number) vs. string wrangling
    formatted_group_names = pd.Series(
        grading_df[naming_dictionary.GRADING_GROUP_ID_COL].apply(lambda i: group_name_fmt.format(i)),
        name=naming_dictionary.MOODLE_GROUP_COL)
    group_scores_df = grading_df[[naming_dictionary.GRADING_GROUP_ID_COL, grading_column] + (
        [feedback_column] if feedback_column else [])].set_index(formatted_group_names)

    merged_df = moodle_df.merge(group_scores_df, on=naming_dictionary.MOODLE_GROUP_COL, how='left')
    merged_df[naming_dictionary.GRADING_GROUP_ID_COL] = merged_df[naming_dictionary.GRADING_GROUP_ID_COL].astype(
        pd.Int64Dtype())

    rows_with_total_score = merged_df[grading_column].notna()

    return merged_df.loc[rows_with_total_score], original_cols


# unused
def join_and_create_overview(output_file: str, naming_dictionary: NamingDictionary, **config):
    updated_df, _ = join_moodle_and_scores(naming_dictionary=naming_dictionary, **config)
    write_table_to_excel(output_file, naming_dictionary.OVERVIEW_TABLE_NAME, updated_df)


def join_and_export(output_file: str, grading_column: str, feedback_column: str, naming_dictionary: NamingDictionary,
                    **config):
    updated_df, original_cols = join_moodle_and_scores(grading_column=grading_column, feedback_column=feedback_column,
                                        naming_dictionary=naming_dictionary, **config)

    # decimal representation wrangling

    stringified_scores_col = stringify_scores_for_moodle(updated_df[grading_column], naming_dictionary)

    changed_rows_idx = updated_df[naming_dictionary.MOODLE_GRADING_COL] != stringified_scores_col

    overridden_rows_idx = changed_rows_idx & updated_df[naming_dictionary.MOODLE_GRADING_COL].notna()

    changed_rows = updated_df.loc[changed_rows_idx]
    if len(changed_rows) > 0:
        if config['verbose']:
            print('Listing changed rows:')
            print_rows_pretty(changed_rows, stringified_scores_col.loc[changed_rows_idx],
                              naming_dictionary=naming_dictionary)
        if config['export_diffs']:
            changed_rows_file = Path(output_file).parent.joinpath('changed_rows.csv').resolve()
            changed_rows.to_csv(changed_rows_file)

    overwritten_rows = updated_df.loc[overridden_rows_idx]
    if len(overwritten_rows) > 0:
        if config['verbose']:
            print('Listing overwritten rows:')
            print_rows_pretty(overwritten_rows, stringified_scores_col.loc[overridden_rows_idx],
                              naming_dictionary=naming_dictionary)

        if config['export_diffs']:
            overwritten_rows_file = Path(output_file).parent.joinpath('overwritten_rows.csv').resolve()
            overwritten_rows.to_csv(overwritten_rows_file)

    # update
    updated_df[naming_dictionary.MOODLE_GRADING_COL] = stringified_scores_col
    if feedback_column:
        updated_df[naming_dictionary.MOODLE_FEEDBACK_COL] = updated_df[feedback_column]

    # select the columns for final export
    if config.get('allow_column_cleaning'):
        export_df = updated_df[[naming_dictionary.MOODLE_MATR_COL] + list(naming_dictionary.MOODLE_IDENTITY_COLS) + [
            naming_dictionary.MOODLE_GRADING_COL, naming_dictionary.MOODLE_FEEDBACK_COL]]
    else:
        export_df = updated_df[original_cols]

    write_moodle_csv(export_df, output_file)
    print(f'Exported {len(changed_rows)} scores into {output_file} ({len(overwritten_rows)} overwritten).')


def configure_parser(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults):
    parser.add_argument('-o', '--output', required=False, help='Path to output csv file.', dest='output_file',
                        default=defaults['DEFAULT_OUTPUT_FILE_NAME'])
    parser.add_argument('-m', '--moodle-grading-worksheet', required=True, help='Path to moodle grading im/export worksheet (a csv file).')
    parser.add_argument('-mgc', '--moodle-grading-column', required=False,
                        help='Name of the grading aspect column in the moodle file.', dest='MOODLE_GRADING_COL',
                        default=naming_dictionary.MOODLE_GRADING_COL)
    parser.add_argument('-mfc', '--moodle-feedback-column', required=False,
                        help='Name of the feedback column in the moodle file.', dest='MOODLE_FEEDBACK_COL',
                        default=naming_dictionary.MOODLE_FEEDBACK_COL)

    parser.add_argument('-g', '--grading-file', required=True, help='Path to grading file with group scores.')

    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-sh', '--grading-sheet', required=False, help='Optionally, excel sheet within grading file.')
    group.add_argument('-t', '--grading-table', required=False, help='Optionally, excel table within grading file.',
                       default=naming_dictionary.GRADING_TABLE_NAME)

    parser.add_argument('-gc', '--grading-column', required=False,
                        help='Name of the grading aspect column in the grading/scores file.',
                        default=naming_dictionary.GRADING_TOTAL_COL)
    parser.add_argument('-gfc', '--feedback-column', required=False,
                        help='Name of the feedback column in the grading/scores file.', default=None)

    parser.add_argument('-gfmt', '--group-name-fmt', required=True,
                        help='Format str to recreate the group names used in moodle group creation (in Python style, e.g., "Group A1 {:02}").')

    parser.add_argument('-acc', '--allow-column-cleaning', required=False, action='store_true', default=False,
                        help='Whether the exported table can be cleaned up. By default, the original columns of the moodle file are preserved, to not support the upload in the grading table upload in the, e.g., assignment activity itself.')
    parser.add_argument('-v', '--verbose', required=False, action='store_true', default=False,
                        help='Whether to produce more verbose output.')
    parser.add_argument('-exd', '--export-diffs', required=False,
                        help='Whether to export tables of changed/overwritten scores (includes initially unset scores).')


class GenGroupUpload(CommandModule):
    module_name = 'gen-group-upload'
    commands = [('scores', configure_parser, join_and_export)]

    @property
    def additional_config(self) -> dict[str, Any]:
        return {'USED_JOIN_COL': 'MOODLE_GROUP_COL',
                'DEFAULT_OUTPUT_FILE_NAME': f'moodle_upload_{datetime.now().strftime("%Y_%m_%d_%H_%M")}.csv'}


if __name__ == '__main__':
    GenGroupUpload().as_program('gen').parse_and_run()



from __future__ import annotations

from collections.abc import Iterator, Iterable
from typing import Any, Self, Callable, TypeVar, Annotated, TypedDict, Literal, Union

import pydantic
import yaml
from pydantic import ConfigDict, TypeAdapter

DEFAULT_QUESTION_DIVISION_LEVELS = ['Assignment', 'Question', 'Subquestion', 'Subsubquestion']

T = TypeVar('T')


class NodeProperties(TypedDict, total=False):
    pts: float | None = None
    mandatory: bool | None = None
    text: str | None = None
    summary: bool | None = None
    skip: bool | None = None
    skip_in_total: bool | None = None
    skip_in_formula: bool | None = None


class GradableNode(pydantic.BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    label: str
    pts: float | None = pydantic.Field(default=None, exclude=True)
    properties: NodeProperties | None = pydantic.Field(default=None)
    parent: GradableNode | None = pydantic.Field(default=None, repr=False, exclude=True)
    children: Annotated[list[GradableNode], pydantic.Field(default_factory=list), pydantic.SkipValidation]

    @property
    def is_leaf(self) -> bool:
        return not self.children

    @property
    def left_siblings(self) -> Iterator[Self]:
        if self.parent:
            for c in self.parent.children:
                if c == self: break
                yield c

    @property
    def right_siblings(self) -> Iterator[Self]:
        if self.parent:
            start_yielding = False
            for c in self.parent.children:
                if start_yielding:
                    yield c
                if c == self: start_yielding = True

    @property
    def siblings(self) -> Iterator[Self]:
        if self.parent:
            yield from (c for c in self.parent.children if c != self)

    # @pydantic.computed_field(repr=False)
    @property
    def ancestry(self) -> Iterator[Self]:
        if self.parent:
            yield from self.parent.ancestry
        yield self

    def get_property(self, k: str) -> Any:
        if self.properties:
            return self.properties.get(k)

    def select_children(self, selector: Callable[[GradableNode], bool]) -> list[Self]:
        return [c for c in self.children if selector(c)]

    @property
    def height(self) -> int:
        if not self.children:
            return 0
        return 1 + max(c.height for c in self.children)

    def iter_level(self, level: int) -> Iterable[GradableNode]:
        if level == 0:
            yield self
        elif level > 0:
            for c in self.children:
                yield from c.iter_level(level - 1)

    def preorder(self, mapper: Callable[[GradableNode], T] = lambda x: x) -> Iterator[T]:
        yield mapper(self) if mapper else self
        if self.children:
            for c in self.children:
                yield from c.preorder(mapper)
                # for c in s.preorder():
                #    yield c

    # depth of zero equals root (self) level
    def cutoff(self, depth: int) -> Self:
        children = [] if depth == 0 else [s.cutoff(depth - 1) for s in self.children]
        return self.__class__(label=self.label, pts=self.pts, info=self.properties, subparts=children)

    def as_nested(self) -> dict[str, dict] | str:
        return {self.label: [c.as_nested() for c in self.children]} if self.children else self.label

    def paths(self) -> list[tuple[Self, ...]]:
        return [(self,) + v for c in self.children for v in c.paths()] if self.children else [(self,)]

    def paths_iter(self, mapper: Callable[[Self], T] | None = lambda x: x) -> Iterator[tuple[T, ...]]:
        res = (mapper(self),) if mapper else (self,)
        if not self.children:
            yield res
        else:
            for c in self.children:
                for p in c.paths_iter(mapper):
                    yield res + p

    def leaves(self) -> list[Self]:
        if not self.children:
            return [self]
        else:
            li = []
            for c in self.children:
                li.extend(c.leaves())
            return li

    def leaves_iter(self, mapper: Callable[[Self], T] = lambda x: x) -> Iterator[T]:
        if not self.children:
            yield mapper(self) if mapper else self
        else:
            for c in self.children:
                yield from c.leaves_iter(mapper)

    @property
    def leaf_count(self) -> int:
        return sum(1 for _ in self.leaves_iter())

    @pydantic.model_validator(mode='after')
    def leaf_or_list(self):
        if not (self.pts is not None or self.children):
            raise ValueError('Leaf nodes need to have points associated to them.')
        return self

    def bubble_up(self, func: Callable[[Self, Iterable[T]], T]) -> T:
        return func(self, (c.bubble_up(func) for c in self.children)) if self.children else func(self, ())

    def trickle_down(self, func: Callable[[Self, T | None], T], drip_depth: int = 0,
                     parent_value: T | None = None) -> None:
        if drip_depth < 0:
            return

        v = func(self, parent_value)
        if self.children:
            for c in self.children:
                c.trickle_down(func, drip_depth - 1, v)

    def complete(self, completion: Callable[[Self, Iterable[Self]], Self] | None = None) -> Self:
        if not completion:
            def completion(n: Self, children: Iterable[Self]) -> Self:
                s = 0
                for c in children:
                    c.parent = n
                    if not c.get_property('skip') and not c.get_property('skip_in_total'):
                        s += c.pts
                # assert n.pts is None or not had_children
                if n.children:
                    n.pts = s
                return n
        return self.bubble_up(completion)

    def make_self_summary(self) -> Self | None:
        return self.__class__(label='Total', pts=self.pts,
                              properties={'skip': True, 'summary': True},
                              children=[], parent=self) if (
                self.children and not self.get_property('skip') and not self.get_property(
            'skip_in_total')) else None

    def add_summaries(self, depth: int = 0) -> None:
        def summarize_children(n: Self, *args):
            if n.children:
                summary = n.make_self_summary()
                if summary:
                    n.children.append(summary)

        self.trickle_down(summarize_children, depth)

    # def collect_ranges(self, start: int = 0) -> Iterator[range]:
    #    if self.is_leaf and (r := self.get_info('summarizes')):
    #        yield range(start, start + r + 1)
    #    else:
    #        leaves_inbetween = 0
    #        for c in self.siblings:
    #            yield from c.collect_ranges(start + leaves_inbetween)
    #            leaves_inbetween += c.leaf_count


GradableNode.model_rebuild()


class GradableNodeFileReprV0(GradableNode):
    model_config = ConfigDict(populate_by_name=True)

    @classmethod
    def from_dict(cls: type[Self], dic) -> Self:
        k, v = next(iter(dic.items()))
        if isinstance(v, list):
            return cls(label=k, children=[cls.from_dict(s) for s in v])
        elif isinstance(v, dict):
            return cls(label=k, **{f: v.pop(f) for f in GradableNode.model_fields.keys() if f in v},
                       properties=(v if v else None))
        else:
            return cls(label=k, pts=v)

    @classmethod
    def __get_validators__(cls):
        yield cls.validate

    @classmethod
    def validate(cls: Self, value: Any) -> Self:
        return cls.from_dict(value)

    @pydantic.model_serializer(mode='wrap')
    def serial(self, srlz: pydantic.SerializerFunctionWrapHandler) -> dict:
        if self.children:
            return {self.label: [srlz(sub) for sub in self.children]}
        elif self.properties:
            return {self.label: srlz({'pts': self.pts}.update(self.properties))}
        else:
            return {self.label: srlz(self.pts)}


class GradableNodeFileReprV1(GradableNode):
    properties: Annotated[NodeProperties | None, pydantic.Field(default=None), pydantic.SkipValidation]
    children: Annotated[
        list[GradableNodeFileReprV1], pydantic.Field(default_factory=list, exclude=True), pydantic.SkipValidation]

    @pydantic.model_validator(mode='wrap')
    @classmethod
    def custom_val(cls, value: Any, nxt: pydantic.ValidatorFunctionWrapHandler):
        assert isinstance(value, dict)

        properties = {}
        for k, v in value.items():
            if k not in GradableNode.model_fields:
                properties[k] = v
        for k in properties:
            value.pop(k)
        if properties:
            value['properties'] = properties

        children = cls.val_children(value.get('children'))
        if children:
            value['children'] = children

        return nxt(value)

    # @pydantic.field_validator('children', mode='wrap')
    @classmethod
    def val_children(cls, value: Any):
        def swallowed(v):
            try:
                return cls.model_validate(v)
            except ValueError as exc:
                if not str(exc).startswith('Circular reference'):
                    raise exc

        if value is not None:
            children = []
            # for c in value:
            #    children.extend(swallowed(c))
            return [swallowed(c) for c in value]

    @pydantic.model_serializer(mode='wrap', when_used='unless-none')
    def custom_ser(self, nxt: pydantic.SerializerFunctionWrapHandler) -> dict:
        # exclude={'children', 'properties'}
        res = {}
        try:
            res = nxt(self)
        except ValueError as exc:
            if 'Circular reference detected' not in str(exc):
                raise exc
        if props := res.pop('properties', {}):
            res.update(props)
        if self.children:
            res['children'] = self._ser_children(self.children, nxt)
        else:
            res['pts'] = self.pts
        return res

    # @pydantic.field_serializer('children', mode='wrap')
    def _ser_children(self, value, nxt: pydantic.SerializerFunctionWrapHandler):
        def swallowed(v):
            try:
                return v.model_dump(exclude_none=True, exclude_unset=True, exclude_defaults=True)
                # return nxt(v) # nxt([v])[0]
            except ValueError as exc:
                if 'Circular reference detected' not in str(exc):
                    raise exc

        if value:
            return [swallowed(c) for c in value]


class GradableSpecBase(pydantic.BaseModel):
    model_config = ConfigDict(populate_by_name=True)
    version: str
    tree_level_names: list[str] = pydantic.Field(alias='tree-level-names',
                                                 default_factory=lambda: DEFAULT_QUESTION_DIVISION_LEVELS)
    tree: GradableNode = pydantic.Field(default_factory=GradableNode)

    def get_level(self, level_name: str, filter: Callable[[GradableNode], bool] = None) -> list[GradableNode]:
        return [n for n in self.tree.iter_level(self.tree_level_names.index(level_name)) if not filter or filter(n)]

    def cutoff(self, depth: int) -> Self:
        return self.__class__(tree=self.tree.cutoff(depth), **{'tree-level-names': self.tree_level_names[:depth + 1]})


class GradableSpecV0(GradableSpecBase):
    version: Literal['v0'] = 'v0'
    tree: GradableNodeFileReprV0 = pydantic.Field(default_factory=GradableNodeFileReprV0)


class GradableSpecV1(GradableSpecBase):
    version: Literal['v1'] = 'v1'
    tree: GradableNodeFileReprV1 = pydantic.Field(default_factory=GradableNodeFileReprV1)


GradableSpec = Annotated[Union[
    Annotated[GradableSpecV0, pydantic.Tag('v0')], Annotated[GradableSpecV1, pydantic.Tag('v1')]], pydantic.Field(
    discriminator='version')]


def write_schema():
    schema = GradableSpecV1.model_json_schema(by_alias=True)
    from pathlib import Path
    p = Path(__file__).parent.joinpath('gradable-spec-schema.yaml')
    print(f'Writing schema to {p}')
    with p.open('w') as f:
        yaml.dump(schema, f)


def load_spec(spec_path: str, custom_completion_func: Callable[[GradableNode, Iterable[
    GradableNode]], GradableNode] | None = None) -> GradableSpec:
    with open(spec_path, 'r') as spec_file:
        spec = yaml.load(spec_file, Loader=yaml.FullLoader)
        gradable_spec = TypeAdapter(GradableSpec).validate_python(spec)
        gradable_spec.tree.complete(custom_completion_func)
        return gradable_spec


def write_spec(spec: GradableSpecV1, spec_path: str):
    with open(spec_path, 'w', encoding='utf-8') as spec_file:
        dump = spec.model_dump(by_alias=True, exclude_none=True, exclude_unset=True)
        yaml.dump(dump, spec_file, Dumper=yaml.Dumper, indent=2)



import argparse
from abc import abstractmethod, ABC
from collections.abc import Callable
from typing import Any, Sequence

from grading_tools.common.defaults import MinimalConfig, get_base_cfg, NamingDictionary

CommandName = str
CommandFunc = Callable[..., None]
ParserRegistration = Callable[[argparse.ArgumentParser, ...], None]


def merge_args(defaults: MinimalConfig, args: argparse.Namespace) -> MinimalConfig:
    nd = defaults['naming_dictionary']
    args_as_dict = vars(args)
    nd_dict = NamingDictionary.model_dump(nd, exclude_defaults=True, exclude_none=True)
    nd_dict.update({k: v for k, v in args_as_dict.items() if k in NamingDictionary.model_fields})
    return defaults | {'naming_dictionary': NamingDictionary.model_validate(nd_dict, strict=False)} | args_as_dict


class CommandModule(ABC):

    @property
    def additional_config(self) -> dict[str, Any]:
        return {}

    @property
    @abstractmethod
    def module_name(self) -> str:
        pass

    @property
    @abstractmethod
    def commands(self) -> list[tuple[CommandName, ParserRegistration, CommandFunc]]:
        pass

    @property
    def module_properties(self) -> dict[str, Any]:
        return {}

    def register_module(self, parser: argparse.ArgumentParser, **defaults) -> None:
        pass

    def register_command_base(self, parser: argparse.ArgumentParser, **defaults) -> None:
        pass

    def __init__(self, default_config: MinimalConfig | None = None) -> None:
        self.default_config: MinimalConfig = default_config if default_config is not None else get_base_cfg()

    def as_top_level(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(self.module_name)
        self.register(parser)
        return parser

    def register(self, parser: argparse.ArgumentParser):
        neat: bool = self.module_properties.get('neat', True)
        force_subparsers: bool = self.module_properties.get('force_subparsers', False)

        self.register_module(parser, **self.default_config)
        if len(self.commands) > 1 or force_subparsers or not neat:
            commands_parser = parser.add_subparsers(title='commands', dest='command', required=True)
            for name, register_command, command in self.commands:
                p = commands_parser.add_parser(name)
                self.register_command_base(p, **self.default_config)
                register_command(p, **self.default_config)

                def with_args(args: argparse.Namespace, cmd=command, cfg=dict(self.default_config)):
                    cmd(**merge_args(cfg, args))

                p.set_defaults(func=with_args)
        else:
            assert len(self.commands) == 1
            (name, register_command, command) = self.commands[0]
            self.register_command_base(parser, **self.default_config)
            register_command(parser, **self.default_config)
            parser.set_defaults(func=lambda args: command(**merge_args(self.default_config, args)))

    def as_program(self, name: str) -> 'CliProgram':
        return CliProgram(name, self.__class__)


class CliProgram:

    def __init__(self, program_name: str, *module_classes: type[CommandModule],
                 default_config: MinimalConfig | None = None):
        self.program_name: str = program_name
        self.module_classes: tuple[type[CommandModule], ...] = module_classes
        self.default_config: MinimalConfig = default_config if default_config is not None else get_base_cfg()

    def make_parser(self) -> argparse.ArgumentParser:
        parser = argparse.ArgumentParser(self.program_name)
        modules_parser = parser.add_subparsers(title='modules', dest='module', required=True)
        for module_cls in self.module_classes:
            module = module_cls(self.default_config)
            module.default_config |= module.additional_config
            module_parser = modules_parser.add_parser(module.module_name)
            module.register(module_parser)
        return parser

    def parse_and_run(self, args: Sequence[str] | None = None) -> None:
        args = self.make_parser().parse_args(args)
        args.func(args)



from grading_tools.common.commands import CliProgram
from grading_tools.common.defaults import MinimalConfig
from grading_tools.spec_manipulation import *
from grading_tools.upload_export_gen import *
from grading_tools.grading_table_gen import *
from grading_tools.misc_table_gen import *
from grading_tools.diagram_gen import *


def mk_cli_program(default_config: MinimalConfig | None = None) -> CliProgram:
    return CliProgram('grading-toolbox', GenGradingTable, EditGradingTable, GenDiagram, GenGroupUpload, GenGradeUpload,
                      GenMiscTable, ComposeFeedback, ConcatSpecs,
                      default_config=default_config)


if __name__ == '__main__':
    mk_cli_program().parse_and_run()



from __future__ import annotations

from typing import Sequence, Iterable

import pandas as pd

from grading_tools.common.defaults import NamingDictionary
from grading_tools.common.gradable_spec import GradableNode, GradableSpecV1
from grading_tools.common.utils import read_generic_table, read_rwthonline

COLUMN_NAME_LEVEL_START = 1
GK_INFO_TABLE_NAME = 'GKInfo'


def equalize(p: tuple, pad_to: int, trim_front: int = 0):
    if trim_front > 0:
        p = p[trim_front:]
    return p + (p[-1],) * max(pad_to - len(p), 0)


def mk_name(path: Sequence[str], start_depth: int = 1) -> str:
    return '-'.join(path[start_depth:])


def mk_node_name(n: GradableNode, start_depth: int = COLUMN_NAME_LEVEL_START) -> str:
    return mk_name(tuple(n.label for n in n.ancestry), start_depth)


def mk_sum_formula_over_ranges(ranges: list[list[GradableNode]],
                               output_table: str,
                               max_points: float | None = None,
                               with_equals: bool = True,
                               upper_label_level: int = COLUMN_NAME_LEVEL_START, **config) -> tuple[str, bool]:
    formula = ''
    for i, local_range in enumerate(ranges):
        s, e = local_range[0], local_range[-1]
        if i > 0:
            formula += '+'
        formula += f'SUM([@[{mk_node_name(s, upper_label_level)}]:[{mk_node_name(e, upper_label_level)}]])'
    if max_points:
        formula = f'MIN({max_points},{formula})'
    if with_equals:
        formula = '=' + formula
    return formula, False


def mk_is_checked_formula(column: str, symbol: str, table: str | None = None):
    column_key = f'[@[{column}]]' if table is None else f'{table}[{column}]'
    # return  f'OR({column_key}="{symbol}",{column_key}="R{symbol}")'
    return f'ISNUMBER(SEARCH("{symbol}",{column_key}))'


def mk_is_checked_condition(column: str | tuple[str, str], symbol: str, table: str | None = None):
    if type(column) == tuple and len(column) == 2:
        column = f'{column[0]}]:[{column[1]}'
    column_key = f'[@[{column}]]' if table is None else f'{table}[{column}]'
    return column_key, f'"*{symbol}*"'


def mk_mandatory_col_disjunction(nodes: Iterable[GradableNode], upper_label_level: int = COLUMN_NAME_LEVEL_START,
                                 **config) -> str | None:
    s = ','.join(
        (mk_is_checked_formula(mk_node_name(n, upper_label_level), config["CHECKBOX_SYMBOL"]) for n in nodes if
         n.get_property('mandatory')))
    return f'OR({s})' if s else None


def mk_mandatory_col_condition(nodes: Iterable[Iterable[GradableNode]], **config) -> str | None:
    s = ','.join((dis for rng in nodes if (dis := mk_mandatory_col_disjunction(rng, **config))))
    return f'AND({s})' if s else None


def mk_marked_formula_over_ranges(ranges: list[list[GradableNode]],
                                  output_table: str, *,
                                  max_points: float | None = None,
                                  with_equals: bool = True,
                                  upper_label_level: int = COLUMN_NAME_LEVEL_START,
                                  naming_dictionary: NamingDictionary = None,
                                  **config) -> tuple[str, bool]:
    mandatory_col_groups = []
    sub_sums = []
    sub_formulas = []
    for i, local_range in enumerate(ranges):
        s, e = local_range[0], local_range[-1]
        start_col = mk_node_name(s, upper_label_level)
        end_col = mk_node_name(e, upper_label_level)
        # cancelled due to excel non-cooperation
        # this formula works if it is entered via the interface..
        # local_formula = f'SUMPRODUCT(IF(IFERROR(SEARCH("{config["CHECKBOX_SYMBOL"]}",[@[{start_col}]:[{end_col}]]),FALSE),1,0),INDEX({GK_INFO_TABLE_NAME}[[{start_col}]:[{end_col}]],1,0))'
        # local_formula = f'SUM(IFERROR(SEARCH("{config["CHECKBOX_SYMBOL"]}",[@[{start_col}]:[{end_col}]]),FALSE) * INDEX({GK_INFO_TABLE_NAME}[[{start_col}]:[{end_col}]],1,0))'
        rg, crit = mk_is_checked_condition((start_col, end_col), config["CHECKBOX_SYMBOL"])
        local_formula = f'SUMIF({rg},{crit},INDEX({naming_dictionary.GRADING_GKINFO_TABLE_NAME}[[{start_col}]:[{end_col}]],1,0))'
        sub_formulas.append(local_formula)
        cns = (mk_node_name(n, upper_label_level) for n in local_range)
        ifs = (
            f'IF({mk_is_checked_formula(c, config["CHECKBOX_SYMBOL"])},INDEX({naming_dictionary.GRADING_GKINFO_TABLE_NAME}[[{c}]],1),0)'
            for c in cns)
        sub_sums.extend(ifs)

        mandatory_col_options = [n for n in local_range if n.get_property('mandatory')]
        if len(mandatory_col_options) > 0:
            mandatory_col_groups.append(mandatory_col_options)

    formula = f'SUM({",".join(sub_formulas)})'

    if max_points:
        formula = f'MIN({max_points},{formula})'
    if len(mandatory_col_groups) > 0:
        s = mk_mandatory_col_condition(mandatory_col_groups, **config)
        if s:
            formula = f'IF({s},{formula},"")'
    if with_equals:
        formula = '=' + formula
    return formula, False


def mk_sum_formula_range(summary_range: Iterable[GradableNode]) -> tuple[list[list[GradableNode]], float]:
    local_max_pts = 0
    ranges = []
    range_start, range_end = (None, None)
    local_range = []
    gap = True
    for j, c in enumerate(summary_range):
        if not c.get_property('bonus') and not c.get_property('skip') and not c.get_property('skip_in_total'):
            local_max_pts += c.pts

        if c.get_property('skip') or c.get_property('skip_in_formula'):
            if not gap:
                assert local_range[0] == range_start
                assert local_range[-1] == range_end
                ranges.append(local_range)
            local_range = []
            range_start = None
            gap = True
        else:
            if gap:
                local_range = []
                range_start = c
                gap = False
        range_end = c
        local_range.append(c)
    if not gap:
        assert local_range[0] == range_start
        assert local_range[-1] == range_end
        ranges.append(local_range)
    return ranges, local_max_pts


def mk_header_info(spec: GradableSpecV1, cutoff_level: str | int | None = None, totals_level: str | int | None = None,
                   **config) -> tuple[
    pd.MultiIndex, list[str], list[float], list[tuple[str, ...]]]:
    if cutoff_level is not None:
        if type(cutoff_level) is str:
            cutoff_level = spec.tree_level_names.index(cutoff_level)
        spec = spec.cutoff(cutoff_level)
        if config['verbose']:
            print(f'Trimmed spec to depth {cutoff_level}')
            print(spec)

    point_tree = spec.tree  # assumed to be `complete()`
    if totals_level is not None:
        if type(totals_level) is str:
            totals_level = spec.tree_level_names.index(totals_level)
        point_tree.add_summaries(totals_level)

    max_pts = sum(n.pts for n in point_tree.leaves_iter() if not n.get_property('bonus') and not n.get_property('skip'))
    pts_list = list(map(lambda n: n.pts, point_tree.leaves_iter()))
    paths = list(point_tree.paths_iter(lambda n: n.label))
    depth = max(map(len, paths))
    paths_list = [equalize(p, depth) for p in paths]
    header_index = pd.MultiIndex.from_tuples(paths_list, names=spec.tree_level_names[:depth])
    column_names_list = [mk_node_name(n) for n in point_tree.leaves_iter()]
    return header_index, column_names_list, pts_list, paths_list


def mk_initialization_df(spec: GradableSpecV1, column_names_list: list[str], **config) -> pd.DataFrame:
    index_columns = config.get('index_columns', config['DEFAULT_INDEX_COLUMNS'])
    df = None
    if (data_df := config.get('data_df')) is not None:
        df = data_df
    elif (data_file := config.get('data_file')) is not None:
        df = read_generic_table(data_file, sheet_name=config.pop('data_sheet', None),
                                table_name=config.pop('data_table', None))
        df = df.set_index(index_columns, drop=True)
    if df is not None:
        keep = {mk_node_name(n) for n in spec.tree.leaves_iter() if not n.get_property('skip')}
        keep = [c for c in df.columns if c in keep]
        initialization_df = df[keep].reindex(columns=column_names_list)
    else:
        if index_file := config.get('index_file'):
            index_file_preset = config.get('index_file_preset', 'manual')
            if index_file_preset == 'rwthonline':
                df = read_rwthonline(index_file)
                nd: NamingDictionary = config.get('naming_dictionary', NamingDictionary())
                rwth_col_names = list(nd.RWTHONLINE_IDENTITY_COLS) + [nd.RWTHONLINE_MATR_COL]
                new_col_names = list(nd.IDENTITY_COLS) + [nd.MATR_COL]
                df = df[rwth_col_names]
                df.columns = new_col_names
            else:
                df = read_generic_table(index_file,
                                        columns=(index_columns if 'index_columns' in config else None),
                                        sheet_name=config.pop('index_sheet', None),
                                        table_name=config.pop('index_table', None))
            data_index = pd.MultiIndex.from_frame(df)
        else:
            data_index = pd.MultiIndex.from_tuples(
                [tuple(i for _ in index_columns) for i in range(config.get('placeholder_row_count', 1))],
                names=index_columns)
        initialization_df = pd.DataFrame(data=None, columns=column_names_list, dtype='Float64').reindex(data_index)
    return initialization_df



import argparse
import os

import numpy as np
import plotly.express as px
import plotly.graph_objects as go
from plotly.subplots import make_subplots

from grading_tools.common.commands import CommandModule
from grading_tools.common.defaults import NamingDictionary
from grading_tools.common.gradable_spec import load_spec
from grading_tools.common.utils import read_generic_table, get_possible_grades


def save_plot(fig: go.Figure, name, **config):
    fig.write_image(os.path.join(config['output'], name), format=config['output_format'])


def gen_overview_figs(grades_file: str,
                      naming_dictionary: NamingDictionary,
                      **config):
    scores_plot_title = config['SCORES_TITLE']
    grades_plot_title = config['GRADES_TITLE']
    df = read_generic_table(grades_file, table_name=config.pop('grades_table', None),
                            sheet_name=config.pop('grades_sheet', None))

    melted_scores = df.melt(id_vars=[naming_dictionary.MATR_COL],
                            value_vars=[naming_dictionary.ASSIGNMENT_SCORE_COL, naming_dictionary.EXAM_SCORE_COL,
                                        naming_dictionary.COURSE_SCORE_COL],
                            var_name='Kind', value_name='Score')

    fig_scores_a = px.histogram(melted_scores, x='Score', template='plotly_white', color='Kind', barmode='group',
                                title=scores_plot_title,
                                nbins=25)
    fig_scores_b = px.histogram(melted_scores, x='Score', template='plotly_white', facet_row='Kind',
                                title=scores_plot_title,
                                nbins=25)
    save_plot(fig_scores_a, f'scores-overview-a.png', **config)
    save_plot(fig_scores_b, f'scores-overview-b.png', **config)

    melted_grades = df.melt(id_vars=[naming_dictionary.MATR_COL],
                            value_vars=[naming_dictionary.EXAM_GRADE_COL, naming_dictionary.COURSE_GRADE_COL],
                            var_name='Kind',
                            value_name='Grade')

    possible_grades = get_possible_grades(absent=False)
    # simply append the other statuses, e.g., "did not show"
    possible_grades.extend(
        (set(df[naming_dictionary.EXAM_GRADE_COL].unique()) | set(
            df[naming_dictionary.COURSE_GRADE_COL].unique())) - set(possible_grades))

    fig_grades_a = px.histogram(melted_grades, x='Grade', template='plotly_white', color='Kind', barmode='group',
                                title=grades_plot_title, category_orders={'Grade': possible_grades})
    fig_grades_b = px.histogram(melted_grades, x='Grade', template='plotly_white', facet_row='Kind',
                                title=grades_plot_title,
                                category_orders={'Grade': possible_grades})

    return fig_grades_a, fig_grades_b


def gen_overview(**config):
    fig_grades_a, fig_grades_b = gen_overview_figs(**config)
    save_plot(fig_grades_a, f'grades-overview-1.png', **config)
    save_plot(fig_grades_b, f'grades-overview-2.png', **config)


def gen_per_question_fig(gradable_spec: str, grading_file: str, naming_dictionary: NamingDictionary, **config):
    spec = load_spec(gradable_spec)
    points_df = read_generic_table(grading_file, table_name=config.pop('grading_table', None),
                                   sheet_name=config.pop('grading_sheet', None))

    for c in points_df.columns:
        if c.startswith('Q') and '-' not in c:
            points_df[c + '-Total'] = points_df[c]
    total_cols = [c for c in points_df.columns if c.startswith('Q') and c.endswith('-Total')]  # ouch
    question_totals = points_df[[naming_dictionary.MATR_COL] + total_cols]

    melted_question_totals = question_totals.melt(id_vars=[naming_dictionary.MATR_COL], var_name='Question',
                                                  value_name='Points')

    total_cols_ordered = [q.label + '-Total' for q in spec.get_level('Question')]
    fig = make_subplots(rows=len(total_cols), subplot_titles=total_cols_ordered, x_title='Points')

    for i, q in enumerate(spec.get_level('Question'), 1):
        col = q.label + '-Total'
        g = go.Histogram(x=melted_question_totals[melted_question_totals['Question'] == col]['Points'], name=col,
                         xbins=dict(start=-0.5, end=int(np.ceil(q.pts + 0.5)), size=1))
        fig.add_trace(g, row=i, col=1)
        fig.update_xaxes(range=[0, q.pts], tick0=0, row=i, col=1)

    fig.update_layout(height=2000, title_text='Points per Question', template='plotly_white')
    return fig


def gen_per_question(**config):
    fig = gen_per_question_fig(**config)
    save_plot(fig, f'points-per-question.png', **config)


def configure_parser_per_question(parser: argparse.ArgumentParser, **defaults):
    parser.add_argument('-gs', '--gradable-spec', required=True, help='Path to the specification of this gradable.')
    parser.add_argument('-gp', '--grading-file', required=True, help='Path to grading file with per-question points.')
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-gpsh', '--grading-sheet', required=False, help='Optionally, excel sheet within grading file.')
    group.add_argument('-gpt', '--grading-table', required=False, help='Optionally, excel table within grading file.')


def configure_parser_overview(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults) -> None:
    parser.add_argument('-gr', '--grades-file', required=True, help='Path to the grading/scores file.', dest='grades_file')
    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-grsh', '--grades-sheet', required=False, help='Optionally, excel sheet within grades file.')
    group.add_argument('-grt', '--grades-table', required=False, help='Optionally, excel table within grades file.',
                       default=naming_dictionary.OVERVIEW_TABLE_NAME)

    parser.add_argument('--assignment-score-column', required=False, help='Column with assignment score.',
                        default=naming_dictionary.ASSIGNMENT_SCORE_COL, dest='ASSIGNMENT_SCORE_COL')
    parser.add_argument('--exam-score-column', required=False, help='Column with exam score.',
                        default=naming_dictionary.EXAM_SCORE_COL, dest='EXAM_SCORE_COL')
    parser.add_argument('--course-score-column', required=False, help='Column with course total score.',
                        default=naming_dictionary.COURSE_SCORE_COL, dest='COURSE_SCORE_COL')
    parser.add_argument('--exam-grade-column', required=False, help='Column with exam grade.',
                        default=naming_dictionary.EXAM_GRADE_COL, dest='EXAM_GRADE_COL')
    parser.add_argument('--course-grade-column', required=False, help='Column with course grade.',
                        default=naming_dictionary.COURSE_GRADE_COL, dest='COURSE_GRADE_COL')


def configure_base_parser(parser: argparse.ArgumentParser) -> None:
    parser.add_argument('-o', '--output', required=False, help='Path to output directory.', default='./')
    parser.add_argument('-of', '--output-format', required=False, choices=['png', 'svg', 'jpg', 'pdf', 'html'],
                        help='Image file format to use.', default='png')


def configure_all(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults) -> None:
    configure_parser_overview(parser, naming_dictionary)
    configure_parser_per_question(parser)


def gen_all(**config):
    gen_overview(**config)
    gen_per_question(**config)


class GenDiagram(CommandModule):
    module_name = 'gen-diagrams'
    commands = [('overview', configure_parser_overview, gen_overview),
                ('per-question', configure_parser_per_question, gen_per_question),
                ('all', configure_all, gen_all)]
    additional_config = {'SCORES_TITLE': 'Scores Overview', 'GRADES_TITLE': 'Grades Overview'}

    def register_command_base(self, parser: argparse.ArgumentParser, **defaults) -> None:
        configure_base_parser(parser)


if __name__ == '__main__':
    GenDiagram().as_program('gen').parse_and_run()



import argparse
from datetime import datetime
from pathlib import Path
from typing import Literal, Any

import pandas as pd

from grading_tools.common.commands import CommandModule
from grading_tools.common.defaults import *
from grading_tools.common.compat_utils import stringify_grades_for_rwthonline
from grading_tools.common.utils import read_generic_table, write_moodle_csv, \
    write_rwthonline_csv, read_rwthonline, read_moodle_excel, get_possible_grades, complete_file_path

OLD_RWTHONLINE_GRADE = 'OLD_RWTHONLINE_GRADE'


def project(df: pd.DataFrame, extra: list[str] = None, naming_dictionary: NamingDictionary = None) -> pd.DataFrame:
    if naming_dictionary is None:
        naming_dictionary = NamingDictionary()
    cols = list(naming_dictionary.IDENTITY_COLS)
    if extra:
        cols.extend(extra)
    else:
        cols.extend([naming_dictionary.EXAM_SCORE_COL, naming_dictionary.COURSE_SCORE_COL,
                     naming_dictionary.COURSE_GRADE_COL])
    return df[cols]



def gen_both(output_file: str, **config):
    grades_df = read_grades_df(**config)
    rwth_export_file = output_file.rsplit('.csv')[0] + '_rwth.csv'
    moodle_export_file = output_file.rsplit('.csv')[0] + '_moodle.csv'
    rwth_export, considered_grades = create_rwthonline_export(grades_df, rwth_file=config.pop('rwth_file', None),
                                                              **config)
    write_rwthonline_csv(rwth_export, rwth_export_file)
    moodle_export, _ = create_moodle_export(considered_grades, moodle_file=config.pop('moodle_file', None), **config)
    write_moodle_csv(moodle_export, moodle_export_file)


def gen_rwth(output_file: str, **config):
    grades_df = read_grades_df(**config)
    rwth_export, _ = create_rwthonline_export(grades_df, rwth_file=config.pop('rwth_file', None), **config)
    write_rwthonline_csv(rwth_export, complete_file_path(output_file, default_file_name=config.pop('DEFAULT_OUTPUT_FILE_NAME', None)))


def gen_moodle(output_file: str, **config):
    considered_grades = read_grades_df(**config)
    moodle_export, _ = create_moodle_export(considered_grades, moodle_file=config.pop('moodle_file', None), **config)
    write_moodle_csv(moodle_export, complete_file_path(output_file, default_file_name=config.pop('DEFAULT_OUTPUT_FILE_NAME', None)))


def read_grades_df(grades_file: str, naming_dictionary: NamingDictionary, **config) -> pd.DataFrame:
    grades_df = read_generic_table(grades_file, sheet_name=config.pop('grades_sheet', None),
                                   table_name=config.pop('grades_table', None))
    grades_df = grades_df.set_index(naming_dictionary.MATR_COL, drop=False)
    grades_df.index.name = 'INDEX'

    # only keep not NA scores
    return grades_df[grades_df[naming_dictionary.COURSE_GRADE_COL].notna()]


def create_moodle_export(scores_df: pd.DataFrame,
                         moodle_file: str = None,
                         used_grading_column: str = None,
                         used_feedback_column: str = None,
                         moodle_grading_column: str = None,
                         moodle_feedback_column: str = None,
                         naming_dictionary: NamingDictionary = NamingDictionary(),
                         **config) -> tuple[pd.DataFrame, pd.DataFrame]:
    if (rwth_file := config.pop('rwth_file', None)) is not None:
        _, considered_grades = create_rwthonline_export(scores_df, rwth_file, **config)
    else:
        considered_grades = scores_df

    moodle_df = read_moodle_excel(moodle_file)

    moodle_df = moodle_df.set_index(naming_dictionary.MOODLE_EXCEL_MATR_COL, drop=False)
    moodle_df.index.name = 'INDEX'
    # only students listed on moodle can receive a grade there
    moodle_export = considered_grades.join(
        moodle_df[
            list(naming_dictionary.MOODLE_EXCEL_IDENTITY_COLS) + [naming_dictionary.MOODLE_EXCEL_MATR_COL,
                                                            naming_dictionary.MOODLE_EXCEL_ID_COL,
                                                            moodle_grading_column,
                                                            moodle_feedback_column]],
        how='inner')
    moodle_export[used_feedback_column] = moodle_export[used_feedback_column].astype(str)

    moodle_changed_rows_idx = (moodle_export[moodle_grading_column] != moodle_export[used_grading_column]) | (
            moodle_export[moodle_feedback_column] != moodle_export[used_feedback_column])
    moodle_changed_rows = project(moodle_export.loc[moodle_changed_rows_idx],
                                  [used_grading_column, moodle_grading_column, used_feedback_column,
                                   moodle_feedback_column], naming_dictionary=naming_dictionary)
    if config['verbose'] and len(moodle_changed_rows) > 0:
        print('The following rows differ in grading value or feedback text (includes initially unset values)')
        print(moodle_changed_rows)

    moodle_overwritten_rows = moodle_export.loc[moodle_changed_rows_idx & (
            moodle_export[moodle_grading_column].notna() | moodle_export[moodle_feedback_column].notna())]
    moodle_overwritten_rows = project(moodle_overwritten_rows,
                                      [used_grading_column, moodle_grading_column, used_feedback_column,
                                       moodle_feedback_column], naming_dictionary=naming_dictionary)
    if len(moodle_overwritten_rows) > 0:
        print('The following rows have overwritten grading value or feedback text')
        print(moodle_overwritten_rows)
        if config['export_diffs']:
            moodle_diffs_file = Path(config['output_file']).parent.joinpath('moodle_overwritten_rows.csv').resolve()
            moodle_overwritten_rows.to_csv(moodle_diffs_file)

    moodle_export[moodle_grading_column] = moodle_export[used_grading_column]
    moodle_export[moodle_feedback_column] = moodle_export[used_feedback_column]
    moodle_export = moodle_export[
        [naming_dictionary.MOODLE_EXCEL_ID_COL,
         naming_dictionary.MATR_COL] + list(naming_dictionary.MOODLE_EXCEL_IDENTITY_COLS) + [moodle_grading_column,
                                                                                       moodle_feedback_column]]

    return moodle_export, considered_grades


## https://wiki-intern.rwth-aachen.de/display/RD/Notenskala+und+Notenvermerke
# X - nicht erschienen
# NZ - nicht zugelassen
# PA - Prüfung abgebrochen
# U - Ungültig/Täuschung
# Die Titel der Spalten "GRADE", "ECTS_GRADE", "REMARK", "DB_Primary_Key_Of_Candidate" und "DB_Primary_Key_Of_Exam" dürfen weder gelöscht, noch umbenannt werden.
#
# Achten Sie darauf, in den Spalten "DB_Primary_Key_Of_Candidate" und "DB_Primary_Key_Of_Exam" keine Einträge zu verändern oder zu löschen.
#
# May be edited:
# DATE_OF_ASSESMENT
# GRADE
# REMARK
def create_rwthonline_export(scores_df: pd.DataFrame, rwth_file: str = None,
                             naming_dictionary: NamingDictionary = NamingDictionary(), **config) -> \
        tuple[pd.DataFrame, pd.DataFrame]:
    rwth_df = read_rwthonline(rwth_file)

    original_rwth_cols = list(rwth_df.columns)
    rwth_df = rwth_df.set_index(naming_dictionary.RWTHONLINE_MATR_COL, drop=False)
    rwth_df.index.name = 'INDEX'
    rwth_df[OLD_RWTHONLINE_GRADE] = rwth_df[naming_dictionary.RWTHONLINE_GRADE_COL]
    rwth_df = rwth_df.drop(naming_dictionary.RWTHONLINE_GRADE_COL, axis='columns')

    non_registered_students_with_grade = scores_df.loc[
        list(set(scores_df[naming_dictionary.MATR_COL]) - set(rwth_df[naming_dictionary.RWTHONLINE_MATR_COL]))]
    if len(non_registered_students_with_grade) > 0:
        print(
            f'The following students are not registered according to the RWTH Online list but received a grade (will {"" if config["skip_rwthonline_registration_filter"] else "not"} be included in output)')
        print(project(non_registered_students_with_grade, naming_dictionary=naming_dictionary))
    registered_students_without_grade = scores_df.loc[
        list(set(rwth_df[naming_dictionary.RWTHONLINE_MATR_COL]) - set(scores_df[naming_dictionary.MATR_COL]))]
    if len(registered_students_without_grade) > 0:
        print('The following students are registered according to the RWTH Online list but are not in the scores table')
        print(project(registered_students_without_grade, naming_dictionary=naming_dictionary))

    # ensure that only students who are registered according to rwth online can receive a grade, unless this is overridden
    how: Literal['inner', 'left'] = 'inner' if not config['skip_rwthonline_registration_filter'] else 'left'
    considered_grades = scores_df.join(rwth_df, how=how)

    existing_rwthonline_grades = considered_grades[OLD_RWTHONLINE_GRADE].notna()
    non_matching_grades_rwth = considered_grades.loc[existing_rwthonline_grades & (
            considered_grades[OLD_RWTHONLINE_GRADE] != considered_grades[naming_dictionary.COURSE_GRADE_COL])]
    if len(non_matching_grades_rwth) > 0:
        print('The following grades differ between the RWTH Online list and provided scores table')
        rwthonline_diffs = project(non_matching_grades_rwth, [OLD_RWTHONLINE_GRADE, naming_dictionary.COURSE_GRADE_COL],
                                   naming_dictionary=naming_dictionary)
        print(rwthonline_diffs)
        if config['export_diffs']:
            rwthonline_diffs_file = Path(config['output_file']).parent.joinpath(
                'rwthonline_overwritten_rows.csv').resolve()
            rwthonline_diffs.to_csv(rwthonline_diffs_file)

    if config['override_grade_with_rwthonline']:
        overwritten_grades = considered_grades.loc[existing_rwthonline_grades]
        if len(existing_rwthonline_grades) > 0:
            print('Overwriting the following grades from the score table with rwthonline entries')
            print(project(overwritten_grades, [OLD_RWTHONLINE_GRADE], naming_dictionary=naming_dictionary))
            considered_grades.loc[existing_rwthonline_grades, naming_dictionary.COURSE_GRADE_COL] = \
                considered_grades.loc[
                    existing_rwthonline_grades, OLD_RWTHONLINE_GRADE]

    rwth_export = rwth_df.join(considered_grades[naming_dictionary.COURSE_GRADE_COL], how='left')
    stringified_grades = stringify_grades_for_rwthonline(rwth_export[naming_dictionary.COURSE_GRADE_COL])
    assert set(stringified_grades) <= set(get_possible_grades(symbols=True, decimal_sep=','))
    rwth_export[naming_dictionary.RWTHONLINE_GRADE_COL] = stringified_grades
    rwth_export = rwth_export[original_rwth_cols]

    return rwth_export, considered_grades


def register_moodle(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults):
    parser.add_argument('-m', '--moodle', required=True, help='Path to moodle excel file.', dest='moodle_file')
    parser.add_argument('-ugc', '--used-grading-column', required=False,
                        default=defaults['DEFAULT_USED_GRADING_COL_NAME'],
                        help='The column in the scores table to use for the moodle grading aspect.')
    parser.add_argument('-ufc', '--used-feedback-column', required=False,
                        default=defaults['DEFAULT_USED_FEEDBACK_COL_NAME'],
                        help='The column in the scores table to use for the feedback of the moodle grading aspect.')
    parser.add_argument('-mgc', '--moodle-grading-column', required=False,
                        help='The name of the grading column in the moodle export.',
                        default=defaults['DEFAULT_MOODLE_GRADING_COL_NAME'])
    parser.add_argument('-mfc', '--moodle-feedback-column', required=False,
                        help='The name of the feedback column in the moodle export.',
                        default=defaults['DEFAULT_MOODLE_FEEDBACK_COL_NAME'])


def register_rwth(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults):
    parser.add_argument('-r', '--rwth', required=True, help='Path to rwth online exam registrations file.',
                        dest='rwth_file')
    parser.add_argument('--skip-rwthonline-registration-filter', required=False, action='store_true', default=False,
                        help='Whether to keep scores from students who do not appear in the rwthonline table.')
    parser.add_argument('--override-grade-with-rwthonline', required=False, action='store_true', default=True,
                        help='Whether to override grades from the scores table with grades from rwthonline.')


def register_base(parser: argparse.ArgumentParser, naming_dictionary: NamingDictionary, **defaults) -> None:
    parser.add_argument('-o', '--output', required=False, help='Path to output csv file.', dest='output_file',
                        default=defaults['DEFAULT_OUTPUT_FILE_NAME'], type=str)

    parser.add_argument('-g', '--grades', required=True, help='Path to the grade file.', dest='grades_file')

    group = parser.add_mutually_exclusive_group(required=False)
    group.add_argument('-sh', '--grades-sheet', required=False, help='Optionally, excel sheet within scores file.')
    group.add_argument('-t', '--grades-table', required=False, help='Optionally, excel table within scores file.',
                       default=naming_dictionary.OVERVIEW_TABLE_NAME)

    parser.add_argument('-v', '--verbose', required=False, action='store_true', default=False,
                        help='Whether to produce more verbose output.')
    parser.add_argument('-exd', '--export-diffs', required=False,
                        help='Whether to export tables of changed/overwritten scores (includes initially unset scores).')


def register_both(*args, **kwargs):
    register_moodle(*args, **kwargs)
    register_rwth(*args, **kwargs)


class GenGradeUpload(CommandModule):
    module_name = "gen-grade-upload"
    commands = [("rwth", register_rwth, gen_rwth),
                ("moodle", register_moodle, gen_rwth),
                ("both", register_both, gen_both)]

    def register_command_base(self, parser: argparse.ArgumentParser, **defaults) -> None:
        register_base(parser, **defaults)

    @property
    def additional_config(self) -> dict[str, Any]:
        nd = self.default_config['naming_dictionary']
        return {'DEFAULT_MOODLE_GRADING_COL_NAME': nd.MOODLE_GRADING_COL,
                'DEFAULT_MOODLE_FEEDBACK_COL_NAME': nd.MOODLE_FEEDBACK_COL,
                'DEFAULT_USED_GRADING_COL_NAME': nd.EXAM_SCORE_COL,
                'DEFAULT_USED_FEEDBACK_COL_NAME': nd.COURSE_GRADE_COL,
                'DEFAULT_OUTPUT_FILE_NAME': f'upload_{datetime.now().strftime("%Y_%m_%d_%H_%M")}.csv'
                }


if __name__ == '__main__':
    GenGradeUpload().as_program('gen-grade-upload').parse_and_run()



import pandas as pd

from grading_tools.common.defaults import NamingDictionary


def stringify_scores_for_moodle(series: pd.Series, naming_dictionary: NamingDictionary):
    return series.apply(lambda x: f'{x:.2f}'.replace('.', naming_dictionary.MOODLE_CSV_DECIMAL_SEP))

def stringify_grades_for_rwthonline(series: pd.Series):
    return series.astype(str).str.replace('.', ',')


import pathlib
from os import PathLike
from typing import Callable

import pandas as pd


def write_table_helper(exwr: pd.ExcelWriter, df: pd.DataFrame, table_name: str, sheet_name: str | None = None,
                       additional_columns: list[dict] | None = None, include_index: bool = True, **kwargs):
    if sheet_name is None:
        sheet_name = table_name

    df.to_excel(exwr, startrow=1, sheet_name=sheet_name, index=include_index, header=False)
    column_definitions = ([{'header': c} for c in df.index.names] if include_index else []) + [{'header': c} for c in
                                                                                               df.columns]
    if additional_columns is not None:
        column_definitions.extend(additional_columns)
    sheet = exwr.sheets[sheet_name]
    sheet.add_table(0, 0, len(df), len(column_definitions) - 1,
                    {'name': table_name, 'columns': column_definitions} | kwargs)
    sheet.autofit()


def write_to_excel(output_file: str | PathLike[str], writer_func: Callable[[pd.ExcelWriter], None],
                   err_on_file_existence: bool = True, **exwr_kwargs):
    print(f'Writing to {output_file}.')
    path = pathlib.Path(output_file)
    if err_on_file_existence:
        if path.resolve().exists():
            raise Exception('FileWouldBeOverwritten')

    path.parent.mkdir(parents=True, exist_ok=True)
    with pd.ExcelWriter(output_file, engine='xlsxwriter', engine_kwargs=dict(options=dict(use_future_functions=True)), **exwr_kwargs) as exwr:
        writer_func(exwr)


def write_table_to_excel(output_file: str | PathLike[str], table_name: str, df: pd.DataFrame,
                         sheet_name: str | None = None,
                         formula_columns: list[dict] | None = None,
                         additional_writer_func: Callable[[pd.ExcelWriter], None] = None,
                         writing_options: dict | None = None, **table_options):
    if writing_options is None:
        writing_options = {}
    if sheet_name is None:
        sheet_name = table_name

    def my_write(exwr):
        write_table_helper(exwr, df, table_name=table_name, sheet_name=sheet_name, additional_columns=formula_columns,
                           last_column=table_options.get('last_column', False))
        if additional_writer_func:
            additional_writer_func(exwr)

    write_to_excel(output_file, my_write, **writing_options)



from collections.abc import Sequence
from typing import TypedDict

import pydantic


# @dataclasses.dataclass(frozen=True)
class NamingDictionary(pydantic.BaseModel):
    model_config = pydantic.ConfigDict(arbitrary_types_allowed=True)

    # moodle

    ## csv

    ## csv structure
    MOODLE_CSV_SEP: str = ','
    MOODLE_CSV_DECIMAL_SEP: str = ','
    MOODLE_QUOTE_CHAR: str = '"'
    MOODLE_NA_CHAR: str = '-'

    ### csv grading table of moodle activity with group submissions

    MOODLE_ID_COL: str = 'ID'
    MOODLE_MATR_COL: str = 'Matrikelnummer'
    MOODLE_FULL_NAME_COL: str = 'Vollständiger Name'
    MOODLE_IDENTITY_COLS: Sequence[str] = ('Vollständiger Name',)
    MOODLE_GROUP_COL: str = 'Gruppe'
    MOODLE_GRADING_COL: str = 'Bewertung'
    MOODLE_FEEDBACK_COL: str = 'Feedback als Kommentar'

    ### csv export of students with group memberships

    MOODLE_GROUPS_CSV_SEP: str = ';'
    MOODLE_GROUPS_QUOTE_CHAR: str = '"'

    MOODLE_GROUPS_IDENTITY_COLS: Sequence[str] = ('Vorname', 'Nachname')
    MOODLE_GROUPS_MATR_COL: str = 'Matrikelnummer'
    MOODLE_GROUPS_GROUPS_COL: str = 'Gruppen'

    ## excel

    ### excel grading table export

    MOODLE_EXCEL_ID_COL: str = 'LMS-ID'
    MOODLE_EXCEL_MATR_COL: str = 'Matrikelnummer'
    MOODLE_EXCEL_IDENTITY_COLS: Sequence[str] = ('Nachname', 'Vorname')

    # rwth online

    ## csv structure

    RWTHONLINE_CSV_SEP: str = ';'
    # unsure here
    # RWTHONLINE_DECIMAL_SEP = ','
    RWTHONLINE_QUOTE_CHAR: str = '"'

    ### exam registrations table

    RWTHONLINE_MATR_COL: str = 'REGISTRATION_NUMBER'
    RWTHONLINE_IDENTITY_COLS: Sequence[str] = ('FAMILY_NAME_OF_STUDENT', 'FIRST_NAME_OF_STUDENT')
    RWTHONLINE_GRADE_COL: str = 'GRADE'

    # exam management

    MATR_COL: str = 'Matr No'
    FIRST_NAME_COL: str = 'First Name'
    LAST_NAME_COL: str = 'Last Name'
    FULL_NAME_COL: str = 'Full Name'
    IDENTITY_COLS: Sequence[str] = (LAST_NAME_COL, FIRST_NAME_COL)
    PARTICIPATED_COL: str = 'Participated'

    ## grading table

    GRADING_SHEET_NAME: str = 'GradingTable'
    GRADING_TABLE_NAME: str = 'GradingTable'
    GRADING_GKINFO_TABLE_NAME: str = 'GKInfo'
    GRADING_INDIVIDUAL_ID_COL: str = MATR_COL
    GRADING_GROUP_ID_COL: str = 'Group ID'
    GRADING_GROUP_NAME_COL: str = 'Group Name'
    GRADING_TOTAL_COL: str = 'Final Points'

    ## overview table

    OVERVIEW_TABLE_NAME: str = 'Overview'
    ADMISSION_COL: str = 'Admitted'
    PASSED_COL: str = 'Passed'
    ASSIGNMENT_SCORE_COL: str = 'Assignment Score'
    EXAM_SCORE_COL: str = 'Exam Score'
    COURSE_SCORE_COL: str = 'Course Score'
    EXAM_GRADE_COL: str = 'Exam Grade'
    COURSE_GRADE_COL: str = 'Course Grade'

    ## moodle groups table

    GROUPS_INFO_TABLE_NAME: str = 'GroupsInfo'

    ### grade calculation
    # TODO maybe move elsewhere
    GRADE_FALLBACK_VALUE: str = 'X'  # did not show


class BaseConfig(pydantic.BaseModel):
    naming_dictionary: NamingDictionary = NamingDictionary()


class MinimalConfig(TypedDict, total=False):
    naming_dictionary: NamingDictionary


def get_base_cfg() -> MinimalConfig:
    return MinimalConfig(naming_dictionary=NamingDictionary())



import pathlib
from collections import defaultdict
from os import PathLike

import openpyxl
import pandas as pd

from grading_tools.common.defaults import *

nd = NamingDictionary()


def complete_file_path(file_path: str, default_file_name: str | None = None, default_extension: str | None = None) -> \
PathLike[str]:
    if default_file_name is None:
        default_file_name = 'output'
    if default_extension is None:
        default_extension = '.xlsx'

    path = pathlib.Path(file_path)
    if path.is_dir():
        if not default_file_name.endswith('.csv') and not default_file_name.endswith('.xlsx'):
            default_file_name += default_extension
        path = path.joinpath(default_file_name)
    return path


def get_possible_grades(numbers: bool = True, absent: bool = True, symbols: bool = False, decimal_sep: str = '.'):
    possible_grades = []
    if numbers:
        for x in range(1, 5):
            possible_grades.extend((f'{x}{decimal_sep}0', f'{x}{decimal_sep}3', f'{x}{decimal_sep}7'))
    possible_grades.append(f'5{decimal_sep}0')
    if absent or symbols:
        possible_grades.append('X')
    if symbols:
        possible_grades.extend(('NZ', 'PA', 'U', 'Q-Q'))
    return possible_grades


def read_excel_table(file_name: str, table_name: str) -> pd.DataFrame:
    wb = None
    try:
        wb = openpyxl.load_workbook(file_name, read_only=False,
                                    data_only=True)  # openpyxl does not have table info if read_only is True; data_only means any functions will pull the last saved value instead of the formula
        sheet, tbl_range = None, None
        for sheetname in wb.sheetnames:  # pulls as strings
            sheet = wb[sheetname]  # get the sheet object instead of string
            if table_name in sheet.tables:  # tables are stored within sheets, not within the workbook, although table names are unique in a workbook
                tbl = sheet.tables[table_name]  # get table object instead of string
                tbl_range = tbl.ref  # something like 'C4:F9'
                break  # we've got our table, bail from for-loop
        if tbl_range is None:
            print(f'Table {table_name} not found')
            return None
        data = sheet[tbl_range]  # returns a tuple that contains rows, where each row is a tuple containing cells
        content = [[cell.value for cell in row] for row in data]  # loop through those row/cell tuples
        header = content[0]  # first row is column headers
        rest = content[1:]  # every row that isn't the first is data
        df = pd.DataFrame(rest, columns=header)
        return df
    finally:
        if wb:
            wb.close()


def read_generic_table(file_name: str, columns: list[str] | None = None, sheet_name: str | None = None,
                       table_name: str | None = None, **kwargs) -> pd.DataFrame:
    print(f'Reading {table_name}/{sheet_name} from to {file_name}.')

    df = None
    if file_name.endswith('.csv'):
        df = pd.read_csv(file_name)
    elif file_name.endswith('.xlsx'):
        if table_name is not None:
            df = read_excel_table(file_name, table_name)
        elif sheet_name is not None:
            df = pd.read_excel(file_name, sheet_name)

    if columns is not None:
        df = df[columns]

    return df


def read_rwthonline(file_name: str) -> pd.DataFrame:
    col_dtypes = defaultdict(lambda: 'str')
    col_dtypes |= {'REGISTRATION_NUMBER': 'Int64'}
    return pd.read_csv(file_name, sep=nd.RWTHONLINE_CSV_SEP, quotechar=nd.RWTHONLINE_QUOTE_CHAR, dtype=col_dtypes)


def read_moodle_csv(file_name: str | PathLike[str]) -> pd.DataFrame:
    return pd.read_csv(file_name, sep=nd.MOODLE_CSV_SEP, quotechar=nd.MOODLE_QUOTE_CHAR,
                       decimal=nd.MOODLE_CSV_DECIMAL_SEP, na_values=[nd.MOODLE_NA_CHAR])


def read_moodle_students_csv(file_name: str | PathLike[str]) -> pd.DataFrame:
    return pd.read_csv(file_name, sep=nd.MOODLE_GROUPS_CSV_SEP, quotechar=nd.MOODLE_GROUPS_QUOTE_CHAR,
                       decimal=nd.MOODLE_CSV_DECIMAL_SEP, na_values=[nd.MOODLE_NA_CHAR])


# for custom grading aspects that don't belong to a moodle activity, e.g., exam scores
def read_moodle_excel(file_name: str | PathLike[str]) -> pd.DataFrame:
    return pd.read_excel(file_name, na_values=[nd.MOODLE_NA_CHAR])


def write_moodle_csv(df: pd.DataFrame, file_path: str | PathLike[str]):
    p = pathlib.Path(file_path)
    print(f'Writing Moodle Upload to {p}.')
    p.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(file_path, index=False, sep=nd.MOODLE_CSV_SEP, quotechar=nd.MOODLE_QUOTE_CHAR,
              decimal=nd.MOODLE_CSV_DECIMAL_SEP,
              encoding='utf-8')


def write_rwthonline_csv(df: pd.DataFrame, file_path: str | PathLike[str]):
    p = pathlib.Path(file_path).resolve()
    print(f'Writing RWTH Online CSV to {p}.')
    p.parent.mkdir(parents=True, exist_ok=True)
    df.to_csv(file_path, index=False, sep=nd.RWTHONLINE_CSV_SEP, quotechar=nd.RWTHONLINE_QUOTE_CHAR, encoding='utf-8')
